from typing import Optional

from figcli.models.role import Role
from figcli.models.run_env import RunEnv
from prompt_toolkit import prompt
from prompt_toolkit.completion import WordCompleter

from figcli.models.defaults.provider import Provider
from figcli.utils.utils import *
from figcli.config import *
import random
import getpass


class Input:
    """
    Contains some generic prompts that are used by multiple different executions paths in figgy
    """

    @staticmethod
    def select_aws_cli_profile() -> str:
        default_value = 'bastion'
        profile = input(
            f'Please input the aws_cli profile name of your first.last_programmatic user in the MGMT account (Default: {default_value}): ') or default_value
        Utils.stc_validate(profile != '', "You must input a valid aws_cli profile")

        return profile

    @staticmethod
    def get_user(provider: str = 'Please input') -> str:
        okta_username = input(f'{provider} username: ')
        Utils.stc_validate(okta_username != '', "You must input a valid OKTA username")

        return okta_username

    @staticmethod
    def get_password(provider: str = 'Please input') -> str:
        okta_password = getpass.getpass(f'{provider} password: ')
        Utils.stc_validate(okta_password != '', "You must input a valid OKTA password")

        return okta_password

    @staticmethod
    def select_role(valid_roles: List[str]) -> Role:
        input_role = None
        while input_role not in valid_roles:
            input_role = prompt(f'What type of user are you? Options are: {valid_roles}: \n -> ',
                                completer=WordCompleter(valid_roles))
            if input_role not in valid_roles:
                print(f"{input_role} is not a valid user type. Please select from: {valid_roles}")

        return Role(input_role)

    @staticmethod
    def select_region() -> str:
        region = None
        while region not in AWS_REGIONS:
            region = prompt('Please input the AWS region to associate figgy with: ',
                            completer=WordCompleter(AWS_REGIONS))

            if region not in AWS_REGIONS:
                print(f"{region} is not a valid AWS Region, please choose from: {AWS_REGIONS}")

        return region

    @staticmethod
    def select_default_account(valid_envs: List[str]) -> RunEnv:
        environment = None
        while environment not in valid_envs:
            environment = prompt(f'Please select a default account. All commands without the specified `--env` '
                                 f'parameter will run against this account. \n Options are: {valid_envs}: \n -> ',
                                 completer=WordCompleter(valid_envs))

            if environment not in valid_envs:
                print(f"{environment} is not a valid environment type. Please select from: {valid_envs}")

        return RunEnv(environment)

    @staticmethod
    def select_enable_colors() -> bool:
        selection = ''
        c = TerminalFactory(Utils.not_windows()).instance().get_colors()
        while selection.lower() != 'y' and selection.lower() != 'n':
            selection = input(f'Enable colored output? (Colors: {c.fg_rd}RED{c.rs} {c.fg_bl}BLUE{c.rs} '
                              f'{c.fg_gr}GREEN{c.rs} <-- If you see weirdness, select N) Y/n?: ')
            selection = selection.lower() if selection != '' else 'y'
        return selection == 'y'

    @staticmethod
    def select_mfa_enabled() -> bool:
        selection = ''
        while selection.lower() != 'y' and selection.lower() != 'n':
            selection = input(f'Use Multi-factor authentication Y/n?: ')
            selection = selection.lower() if selection != '' else 'y'

        return selection == 'y'

    @staticmethod
    def select_auto_mfa() -> bool:
        return Input.y_n_input('Would you like to save your MFA secret to your keychain and have Figgy generate '
                               f'one time pass codes on your behalf? ', default_yes=False)

    @staticmethod
    def get_mfa_secret() -> str:
        print(f"You have selected that you would like to have `{CLI_NAME}` auto-generate one-time pass codes for you. "
              f"This is going to save you a lot of time, but we'll need to save your MFA secret to your OS keychain.\n"
              f"\nYour keychain secret is a text string that looks something like this: `LYV5Z1SNBM4KKUZO`. This "
              f"is the text representation of the QR Code you would would scan. \n\n")
        secret = input('Please input your MFA secret: -> ')
        return secret

    @staticmethod
    def select_report_errors() -> bool:
        selection = ''
        while selection.lower() != 'y' and selection.lower() != 'n':
            selection = input(f'Help make Figgy better. turn on anonymous error reporting?  Y/n?: ')
            selection = selection.lower() if selection != '' else 'y'
        return selection == 'y'

    @staticmethod
    def select_usage_tracking() -> bool:
        selection = ''
        while selection.lower() != 'y' and selection.lower() != 'n':
            selection = input(f'Help make Figgy better. Turn on 100% anonymous usage reporting Y/n?: ')
            selection = selection.lower() if selection != '' else 'y'
        return selection == 'y'

    @staticmethod
    def select_provider() -> Provider:
        selection = Provider.UNSELECTED
        completer = WordCompleter(words=Provider.names())
        while selection is Provider.UNSELECTED:
            selection = prompt(f'Please select an authentication provider. Options are: {Provider.names()}: ',
                               completer=completer)
            selection = Provider.new(selection)

        return selection

    @staticmethod
    def get_mfa(display_hint: bool = False, color: Optional[Color] = None) -> str:
        if display_hint and random.randint(0, 10) == 10:
            blue = color.fg_bl if color else ''
            rs = color.rs if color else ''
            print(f"{blue}Hint:{rs} Tired of typing in your MFA? Consider saving your MFA secret to your keychain and "
                  f"let {CLI_NAME} securely auto-generate tokens for you. \n"
                  f"{blue}More info:{rs} http://figgy.dev/docs/getting-started/install.html\n\n")

        mfa = input('Please input the MFA associated with your user: ')
        Utils.stc_validate(mfa != '', "You must input a valid mfa")

        return mfa

    @staticmethod
    def select_kms_key(valid_keys: List[str]) -> str:
        return Input.select("Select an encryption key: ", valid_options=valid_keys)

    @staticmethod
    def select_run_env(valid_envs: List[str]) -> RunEnv:
        input_env = prompt(f'Select a RunEnvironment: {valid_envs}: ',
                           completer=WordCompleter(valid_envs))
        Utils.stc_validate(input_env in valid_envs,
                           f"{input_env} is not a valid Run Environment. Please select from: {valid_envs}")

        return RunEnv(input_env)

    @staticmethod
    def is_secret() -> bool:
        return Input.y_n_input("Is this value a secret?", default_yes=False)

    @staticmethod
    def input(message: str, completer=None, default: str = None, optional: bool = False) -> str:
        loop = True
        message = f'\n{message}\n  -> ' if len(message) > 15 else f'\n{message} -> '

        while loop:
            if completer:
                result = prompt(message, completer=completer)
            elif default:
                result = prompt(message, default=default)
            else:
                result = input(message)

            if optional:
                return result

            loop = not result or result.strip() == ''

        return result

    @staticmethod
    def y_n_input(message: str, default_yes: bool = True, invalid_no=False) -> bool:
        """
        Returns True if user selects 'y', or False if user select 'N'
        :param message: Message to prompt the user with.
        :param default_yes: Make the user's default option 'Y'?
        :param invalid_no: If the user enters invalid input, assume a selection of 'N'
        :return: True/False based on user input
        """

        selection = ''
        default_compare = 'y' if default_yes else 'n'
        default_prompt = '(Y/n)' if default_yes else '(y/N)'
        prompt_msg = f'\n{message} \n {default_prompt}: -> '

        while selection.lower() != 'y' and selection.lower() != 'n':
            selection = prompt(prompt_msg, completer=WordCompleter(['y', 'n'])).strip().lower()
            selection = selection.lower() if selection != '' else default_compare

            if selection != 'y' and selection != 'n' and invalid_no:
                return False

        print()
        return True if selection == 'y' else False

    @staticmethod
    def select(message: str, valid_options: List[str], default: str = None) -> str:
        """
        Returns the user's selection based on the provided valid options.
        :param message: Message to prompt the user with.
        :param valid_options: List of valid options to accept.
        :param default: Set the prompt's default to this
        :return: Selected option
        """
        selection = ''
        msg = f'{message}\nOptions: {valid_options}\n -> ' if len(valid_options) < 10 else f'{message} -> '
        while not selection:
            if default:
                selection = prompt(msg, completer=WordCompleter(words=valid_options, match_middle=True), default=default)
            else:
                selection = prompt(msg, completer=WordCompleter(words=valid_options))

            if selection not in valid_options:
                selection = ''

        print()
        return selection



from boto3.resources.base import ServiceResource

from figcli.commands.config.list import List as FigList
from figcli.commands.config.audit import Audit
from figcli.commands.config.browse import Browse
from figcli.commands.config.cleanup import Cleanup
from figcli.commands.config.delete import Delete
from figcli.commands.config.dump import Dump
from figcli.commands.config.edit import Edit
from figcli.commands.config.generate import Generate
from figcli.commands.config.promote import Promote
from figcli.commands.config.restore import Restore
from figcli.commands.config.share import *
from figcli.commands.config.sync import *
from figcli.commands.config.validate import Validate
from figcli.commands.config_context import ConfigContext
from figcli.commands.factory import Factory
from figcli.svcs.kms import KmsSvc
from figcli.svcs.auth.session_manager import SessionManager
from figcli.views.rbac_limited_config import RBACLimitedConfigView
from figcli.utils.utils import *


class ConfigFactory(Factory):
    """
    This factory is used to initialize and return all different types of CONFIG commands. For other resources, there
    would hypothetically be other factories.
    """

    def __init__(self, command: frozenset, context: ConfigContext, ssm: SsmDao, cfg: ConfigDao, kms: KmsSvc,
                 s3_resource: ServiceResource, colors_enabled: bool, config_view: RBACLimitedConfigView,
                 session_manager: SessionManager):

        self._command: frozenset = command
        self._config_context: ConfigContext = context
        self._ssm: SsmDao = ssm
        self._config: ConfigDao = cfg
        self._kms: KmsSvc = kms
        self._colors_enabled: bool = colors_enabled
        self._config_view = config_view
        self._s3_resource: ServiceResource = s3_resource
        self._utils = Utils(colors_enabled)
        self._args = context.args
        self._config_completer = self._config_view.get_config_completer()
        self._session_manager = session_manager

    def instance(self):
        return self.get(self._command)

    def get(self, command: frozenset):
        if command == sync:
            return Sync(self._ssm, self._config, self._colors_enabled, self._config_context, self.get(get),
                        self.get(put))
        elif command == cleanup:
            return Cleanup(self._ssm, self._config, self._config_context, self._config_completer,
                           self._colors_enabled, self.get(delete), args=self._args)
        elif command == put:
            return Put(self._ssm, self._colors_enabled, self._config_context, self._config_view, self.get(get))
        elif command == delete:
            return Delete(self._ssm, self._config, self._config_context, self._colors_enabled,
                          self._config_completer)
        elif command == get:
            return Get(self._ssm, self._config_completer, self._colors_enabled, self._config_context)
        elif command == share:
            return Share(self._ssm, self._config, self._config_completer, self._colors_enabled, self._config_context)
        elif command == list_com:
            return FigList(self._config_view, self._config_completer, self._colors_enabled,
                           self._config_context, self.get(get))
        elif command == browse:
            return Browse(self._ssm, self._colors_enabled, self._config_context, self.get(get),
                          self.get(delete), self._config_view)
        elif command == audit:
            return Audit(self._ssm, self._config, self._config_completer, self._colors_enabled, self._config_context)
        elif command == dump:
            return Dump(self._ssm, self._config_completer, self._colors_enabled, self._config_context)
        elif command == restore:
            return Restore(self._ssm, self._kms, self._config, self._colors_enabled, self._config_context,
                           self._config_completer, self.get(delete))
        elif command == promote:
            return Promote(self._ssm, self._config_completer, self._colors_enabled,
                           self._config_context, self._session_manager)
        elif command == edit:
            return Edit(self._ssm, self._colors_enabled, self._config_context, self._config_view, self._config_completer)
        elif command == generate:
            return Generate(self._colors_enabled, self._config_context)
        elif command == validate:
            return Validate(self._ssm, self._colors_enabled, self._config_context)
        else:
            self._utils.error_exit(f"{Utils.get_first(command)} is not a valid command. You must select from: "
                                   f"[{CollectionUtils.printable_set(config_commands)}]. Try using --help for more info.")



from enum import Enum
from typing import List


class Provider(Enum):
    OKTA = "okta"
    GOOGLE = "google"
    # MICROSOFT = "microsoft"
    # GENERIC_SAML = "saml"
    AWS_BASTION = "bastion"
    PROFILE = "profile"
    UNSELECTED = 0

    @staticmethod
    def sso_providers() -> List["Provider"]:
        return [Provider.OKTA, Provider.GOOGLE]

    @staticmethod
    def names():
        return [provider.name for provider in Provider]

    @staticmethod
    def values():
        return [provider.value for provider in Provider]

    @staticmethod
    def new(name: str):
        if name not in Provider.names():
            return Provider.UNSELECTED
        else:
            return Provider[name]



import json
import logging
import os

import boto3

from botocore.errorfactory import ClientError
from botocore.exceptions import NoCredentialsError, ParamValidationError
from json import JSONDecodeError
from figcli.config import *
from figcli.models.assumable_role import AssumableRole
from figcli.models.defaults.defaults import CLIDefaults
from figcli.svcs.auth.okta.okta import Okta
from figcli.svcs.auth.provider.session_provider import SessionProvider
from figcli.utils.utils import Utils, InvalidSessionError

log = logging.getLogger(__name__)


class SessionManager:
    _MAX_ATTEMPTS = 5

    def __init__(self, defaults: CLIDefaults, session_provider: SessionProvider):
        self._sts = boto3.client('sts')
        self._utils = Utils(defaults.colors_enabled)
        self._defaults = defaults
        self.session_provider: SessionProvider = session_provider

    @Utils.trace
    def get_session(self, assumable_role: AssumableRole, prompt: bool, exit_on_fail=True) -> boto3.Session:
        """
        Creates a session in the specified ENV for the target role from a SAML assertion returned by OKTA authentication.
        Args:
            assumable_role: AssumableRole - The role to be leveraged to authenticate this session
            prompt: If prompt is set, we will not use a cached session and will generate new sessions for okta and mgmt.
            exit_on_fail: Exit the program if this session hydration fails.

        returns: Hydrated session for role + account that match the specified one in the provided AssumableRole
        """
        return self.session_provider.get_session(assumable_role, prompt, exit_on_fail=exit_on_fail)


import pexpect
from figcli.test.cli.config import *
from figcli.test.cli.figgy import FiggyTest
from figcli.config import *
from figcli.utils.utils import *
import os


class DataConfigure(FiggyTest):

    def __init__(self):
        super().__init__(None)

    def run(self):
        print(f"Testing `{CLI_NAME} --{Utils.get_first(configure)}`")
        child = pexpect.spawn(f'{CLI_NAME} --{Utils.get_first(configure)} --skip-upgrade', timeout=3, encoding='utf-8')
        user_name = os.environ.get(SNAGBOT_USER_ENV_KEY)
        password = os.environ.get(SNAGBOT_PASSWORD_ENV_KEY)

        child.expect('.*What type of user.*')
        child.sendline('data')
        child.expect('.*Enable.* colored output.*')
        child.sendline('n')
        child.expect('.*Please input OKTA username.*')
        child.sendline(user_name)
        child.expect('.*Please input OKTA password.*')
        child.sendline(password)
        child.expect('.*default account.*')
        child.sendline(DEFAULT_ENV)
        child.expect('.*figgy successfully configured..*')



import os
import sys
import glob
import readline


# Got this from: https://gist.github.com/iamatypeofwalrus/5637895

class Completer(object):
    """
    A tab completer that can either complete from
    the filesystem or from a list.

    Partially taken from:
    http://stackoverflow.com/questions/5637124/tab-completion-in-pythons-raw-input
    """

    def pathCompleter(self, text, state):
        """
        This is the tab completer for systems paths.
        Only tested on *nix systems
        """
        line = readline.get_line_buffer().split()

        # replace ~ with the user's home dir. See https://docs.python.org/2/library/os.path.html
        if '~' in text:
            text = os.path.expanduser('~')

        # autocomplete directories with having a trailing slash
        if os.path.isdir(text):
            text = text + '/'

        value = [x for x in glob.glob(text + '*')][state]
        if os.path.isdir(value):
            return value + "/"
        else:
            return value

        # return [x for x in glob.glob(text + '*')][state]

    def createListCompleter(self, ll):
        """
        This is a closure that creates a method that autocompletes from
        the given list.

        Since the autocomplete function can't be given a list to complete from
        a closure is used to create the listCompleter function with a list to complete
        from.
        """

        def listCompleter(text, state):
            line = readline.get_line_buffer()

            if not line:
                return [c + " " for c in ll][state]

            else:
                return [c + " " for c in ll if c.startswith(line)][state]

        self.listCompleter = listCompleter



from figcli.commands.config_context import ConfigContext
from figcli.commands.types.config import ConfigCommand
from figcli.config import *
from figcli.data.dao.config import ConfigDao
from figcli.data.dao.ssm import SsmDao
from figcli.extras.key_utils import KeyUtils
from figcli.svcs.observability.anonymous_usage_tracker import AnonymousUsageTracker
from figcli.svcs.observability.version_tracker import VersionTracker
from figcli.utils.utils import Utils


class Validate(ConfigCommand):

    def __init__(self, ssm_init: SsmDao, colors_enabled: bool, context: ConfigContext):
        super().__init__(validate, colors_enabled, context)
        self._ssm = ssm_init
        self._config_path = context.ci_config_path if context.ci_config_path else Utils.find_figgy_json()
        self._utils = Utils(colors_enabled)
        self._replication_only = context.replication_only
        self._errors_detected = False
        self.example = f"{self.c.fg_bl}{CLI_NAME} config {self.command_printable} " \
                       f"--env dev --config /path/to/config{self.c.rs}"
        self._FILE_PREFIX = "file://"

    def _validate(self):
        missing_key = False
        config = self._utils.get_ci_config(self._config_path)
        shared_names = set(self._utils.get_config_key_safe(SHARED_KEY, config, default=[]))
        repl_conf = self._utils.get_config_key_safe(REPLICATION_KEY, config, default={})
        repl_from_conf = self._utils.get_config_key_safe(REPL_FROM_KEY, config, default={})
        merge_conf = self._utils.get_config_key_safe(MERGE_KEY, config, default={})
        config_keys = set(self._utils.get_config_key_safe(CONFIG_KEY, config, default=[]))
        namespace = self._utils.get_namespace(config)
        all_names = KeyUtils.find_all_expected_names(config_keys, shared_names, merge_conf, repl_conf,
                                                     repl_from_conf, namespace)

        all_params = self._ssm.get_all_parameters([namespace])

        all_param_names = []
        for param in all_params:
            all_param_names.append(param['Name'])

        print()
        for name in all_names:
            if name not in all_param_names:
                print(f"Fig missing from {self.c.fg_yl}{self.run_env}{self.c.rs} environment"
                      f" Parameter Store: {self.c.fg_yl}{name}{self.c.rs}")
                missing_key = True
            else:
                print(f"Fig found in ParameterStore: {self.c.fg_bl}{name}{self.c.rs}.")

        if missing_key:
            print("\n\n")
            self._utils.error_exit(f"{MISSING_PS_NAME_MESSAGE}")
        else:
            print(f"{self.c.fg_gr}Success! All figs have been located in the {self.c.rs}{self.c.fg_bl}{self.run_env}"
                  f"{self.c.rs} ParameterStore!")

    @VersionTracker.notify_user
    @AnonymousUsageTracker.track_command_usage
    def execute(self):
        self._validate()



from figcli.utils.secrets_manager import SecretsManager
from figcli.utils.utils import Utils
from cryptography.fernet import Fernet
from figcli.config.constants import KEYCHAIN_ENCRYPTION_KEY, DEFAULT_ENCRYPTION_KEY

"""
Largely taken from simple-crypt, but removing pycrypto requirement, instead using pycryptodome.
Slight tuning to account for faster encrypt/decrypt speeds. Remember, Figgy _only_ deals with temporary
access keys. Nothing in the vault should be valid longer than 12 hours from the time it was originally added
to the vault.
"""


class FiggyVault:

    def __init__(self, keychain_enabled=True):
        """
        keychain_enabled: Stores the encyrption key in the user's keychain. This will be disabled for Sandbox
        sessions to simlify the user experience.
        """
        if keychain_enabled:
            encryption_key = SecretsManager.get_password(KEYCHAIN_ENCRYPTION_KEY)
            if not encryption_key:
                Utils.wipe_vaults()
                encryption_key: str = Fernet.generate_key().decode()
                SecretsManager.set_password(KEYCHAIN_ENCRYPTION_KEY, encryption_key)
        else:
            encryption_key = DEFAULT_ENCRYPTION_KEY

        self.fernet = Fernet(encryption_key)

    def encrypt(self, data):
        """
        Encrypt some data.  Input can be bytes or a string (which will be encoded
        using UTF-8).

        @param data: The data to be encrypted.

        @return: The encrypted data, as bytes.
        """
        return self.fernet.encrypt(self._str_to_bytes(data))

    def decrypt(self, data):
        """
        Decrypt some data.  Input must be bytes.

        @param data: The data to be decrypted, typically as bytes.

        @return: The decrypted data, as bytes.  If the original message was a
        string you can re-create that using `result.decode('utf8')`.
        """
        return self.fernet.decrypt(data)

    @staticmethod
    def _str_to_bytes(data):
        u_type = type(b''.decode('utf8'))
        if isinstance(data, u_type):
            return data.encode('utf8')
        return data



import base64
import logging
import re
import xml.etree.ElementTree as ET
from abc import ABC
from json import JSONDecodeError
from typing import List
from figcli.config import *
from figcli.input import Input
from figcli.models.assumable_role import AssumableRole
from figcli.models.defaults.defaults import CLIDefaults
from figcli.models.role import Role
from figcli.models.run_env import RunEnv
from figcli.models.sso.okta.okta_primary_auth import OktaPrimaryAuth, OktaSession
from figcli.models.sso.okta.okta_session_auth import OktaSessionAuth
from figcli.svcs.cache_manager import CacheManager
from figcli.svcs.auth.okta.okta import Okta, InvalidSessionError
from figcli.svcs.auth.provider.sso_session_provider import SSOSessionProvider
from figcli.svcs.vault import FiggyVault
from figcli.utils.secrets_manager import SecretsManager
from figcli.utils.utils import Utils

log = logging.getLogger(__name__)


class OktaSessionProvider(SSOSessionProvider, ABC):
    _SESSION_CACHE_KEY = 'session'

    def __init__(self, defaults: CLIDefaults):
        super().__init__(defaults)
        keychain_enabled = defaults.extras.get(DISABLE_KEYRING) is not True
        vault = FiggyVault(keychain_enabled=keychain_enabled)
        self._cache_manager: CacheManager = CacheManager(file_override=OKTA_SESSION_CACHE_PATH, vault=vault)
        self._saml_cache: CacheManager = CacheManager(file_override=SAML_SESSION_CACHE_PATH, vault=vault)

    def _write_okta_session_to_cache(self, session: OktaSession) -> None:
        self._cache_manager.write(self._SESSION_CACHE_KEY, session)

    def _get_session_from_cache(self) -> OktaSession:
        last_write, session = self._cache_manager.get(self._SESSION_CACHE_KEY)
        return session

    def get_sso_session(self, prompt: bool = False) -> Okta:
        """
        Pulls the last okta session from cache, if cache doesn't exist, generates a new session and writes it to cache.
        From this session, the OKTA SVC is hydrated and returned.
        Args:
            prompt: If supplied, will never get session from cache.

        Returns: Initialized Okta service.
        """
        count = 0
        while True:
            try:
                if prompt:
                    raise InvalidSessionError("Forcing new session due to prompt.")

                cached_session = self._get_session_from_cache()
                if not cached_session:
                    raise InvalidSessionError

                okta = Okta(OktaSessionAuth(cached_session))
                return okta
            except (FileNotFoundError, InvalidSessionError, JSONDecodeError, AttributeError) as e:
                try:
                    password = SecretsManager.get_password(self._defaults.user)
                    if self._defaults.mfa_enabled:
                        color = Utils.default_colors() if self._defaults.colors_enabled else None
                        mfa = SecretsManager.generate_mfa(self._defaults.user) if self._defaults.auto_mfa else\
                                Input.get_mfa(display_hint=True, color=color)
                    else:
                        mfa = None
                    primary_auth = OktaPrimaryAuth(self._defaults, password, mfa)
                    self._write_okta_session_to_cache(primary_auth.get_session())
                    return Okta(primary_auth)
                except InvalidSessionError as e:
                    count += 1
                    prompt = True
                    log.error(f"Caught error when authing with OKTA & caching session: {e}. ")
                    if count > 3:
                        Utils.stc_error_exit("Unable to autheticate with OKTA with your provided credentials. Perhaps your"
                                         f"user, password, or MFA changed? Try reunning `{CLI_NAME} --configure` again.")
                    # self._defaults = self._setup.basic_configure(configure_provider=self._defaults.provider_config is None)

    @Utils.trace
    def get_saml_assertion(self, prompt: bool = False) -> str:
        """
        Lookup OKTA session from cache, if it's valid, use it, otherwise, generate new assertion with MFA
        Args:
            prompt: Used for forcing prompts of username / password and always generating a new assertion
            force_new: Forces a new session, abandons one from cache
        """
        invalid_session = True
        okta = self.get_sso_session(prompt)
        failure_count = 0
        while invalid_session:
            try:
                assertion = okta.get_assertion()
            except InvalidSessionError as e:
                if failure_count > 0:
                    print(e)
                    print("Authentication failed with SSO provider, please reauthenticate"
                          " Likely invalid MFA or Password?\r\n")
                    failure_count += 1

                log.info(f"GOT INVALID SESSION: {e}")
                user = self._get_user(prompt)
                password = self._get_password(user, prompt=prompt, save=True)

                if self._defaults.mfa_enabled:
                    color = Utils.default_colors() if self._defaults.colors_enabled else None
                    mfa = SecretsManager.generate_mfa(user) if self._defaults.auto_mfa else \
                        Input.get_mfa(display_hint=True, color=color)
                else:
                    mfa = None

                primary_auth = OktaPrimaryAuth(self._defaults, password, mfa)

                try:
                    log.info("Trying to write session to cache...")
                    self._write_okta_session_to_cache(primary_auth.get_session())
                except InvalidSessionError as e:
                    log.info(f"Got invalid session: {e}")
                    return self.get_saml_assertion(prompt=True)
                else:
                    return self.get_saml_assertion(prompt=True)
            else:
                assertion = base64.b64decode(assertion).decode('utf-8')
                self._saml_cache.write(SAML_ASSERTION_CACHE_KEY, assertion)
                return assertion

    def get_assumable_roles(self) -> List[AssumableRole]:
        return self._cache_manager.get_val_or_refresh('assumable_roles', refresher=self.__lookup_roles)

    def __lookup_roles(self) -> List[AssumableRole]:
        assertion = self.get_saml_assertion(prompt=False)
        root = ET.fromstring(assertion)
        prefix_map = {"saml2": "urn:oasis:names:tc:SAML:2.0:assertion"}
        role_attribute = root.find(".//saml2:Attribute[@Name='https://aws.amazon.com/SAML/Attributes/Role']",
                                   prefix_map)

        # SAML arns should look something like this:
        # arn:aws:iam::106481321259:saml-provider/okta,arn:aws:iam::106481321259:role/figgy-dev-data
        pattern = r'^arn:aws:iam::([0-9]+):saml-provider/(\w+),arn:aws:iam::.*role/(\w+-(\w+)-(\w+))'
        assumable_roles: List[AssumableRole] = []
        for value in role_attribute.findall('.//saml2:AttributeValue', prefix_map):
            result = re.search(pattern, value.text)
            unparsable_msg = f'{value.text} is of an invalid pattern, it must match: {pattern} for figgy to ' \
                             f'dynamically map account_id -> run_env -> role for OKTA users.'
            if not result:
                Utils.stc_error_exit(unparsable_msg)

            result.groups()
            account_id, provider_name, role_name, run_env, role = result.groups()

            if not account_id or not run_env or not role_name or not role:
                Utils.stc_error_exit(unparsable_msg)
            else:
                assumable_roles.append(AssumableRole(account_id=account_id,
                                                     role=Role(role, full_name=role_name),
                                                     run_env=RunEnv(run_env),
                                                     provider_name=provider_name,
                                                     profile=None))
        return assumable_roles

    def cleanup_session_cache(self):
        self._write_okta_session_to_cache(OktaSession("", ""))



import argparse
import logging
from figcli.config import *
from figcli.models.assumable_role import AssumableRole
from figcli.models.run_env import RunEnv
from figcli.models.role import Role
from figcli.utils.utils import Utils
from typing import Optional, Dict, Union, List, Set

logger = logging.getLogger(__name__)


class FiggyContext:
    """
    Contains contextual data required for building commands based on CLI input.
    """

    def __init__(self, colors_enabled: bool, resource: frozenset, command: frozenset,
                 run_env: RunEnv, role: AssumableRole, args: argparse.Namespace):
        self.colors_enabled = colors_enabled
        self.command: frozenset = command
        self.resource: frozenset = resource

        # This enables us to have "commands" without resources. Like `figgy login` instead of `figgy login login`.
        # Makes things a bit more flexible.
        self.command = self.command if command else resource

        self.args = args
        self.run_env: RunEnv = run_env
        self.selected_role: AssumableRole = role
        self.role: Role = self.selected_role.role if self.selected_role else None
        self.ci_config_path = Utils.attr_if_exists(config, args)
        self.from_path = Utils.attr_if_exists(from_path, args)
        self.out_file = Utils.attr_if_exists(out, args)
        self.query_prefix = Utils.attr_if_exists(prefix, args)
        self.service = Utils.attr_if_exists(service, args)
        self.profile = Utils.attr_if_exists(profile, args)

        # Flags like --prompt that are unset or set to true
        self.repl = Utils.is_set_true(replication_only, args)
        self.manual = Utils.is_set_true(manual, args)
        self.point_in_time = Utils.is_set_true(point_in_time, args)
        self.all_profiles = Utils.is_set_true(all_profiles, args)
        self.skip_upgrade = Utils.is_set_true(skip_upgrade, args)
        self.configure = Utils.is_set_true(configure, args)
        self.version = Utils.is_set_true(version, args)

        logger.info(self.__dict__)

    def has_optional_arguments(self, argument: frozenset):
        return Utils.is_set_true(argument, self.args)

    def find_matching_optional_arguments(self, arguments: List[frozenset]) -> Set[frozenset]:
        optional_args = set()
        for arg in arguments:
            if Utils.is_set_true(arg, self.args):
                optional_args.add(arg)

        return optional_args



from typing import Optional

from figcli.models.defaults.defaults import CLIDefaults
from figcli.models.run_env import RunEnv
from figcli.models.role import Role
from figcli.commands.command_context import CommandContext


class IAMContext(CommandContext):
    """
    All `iam` resource commands require this context for general use. Config commands operate differently based on
    Role and RunEnv
    """
    def __init__(self, run_env: RunEnv, role: Role, colors_enabled: bool, resource: frozenset,
                 defaults: Optional[CLIDefaults]):
        super().__init__(run_env, resource, defaults=defaults)
        self.role: Role = role
        self.colors_enabled = colors_enabled



from sty import fg

from figcli.config.style.color import Color
from figcli.config.style.palette import Palette
from figcli.config.style.terms.term import Term


class DefaultTerm(Term):
    TERM_PROGRAM = 'Unknown'
    BLUE = Palette.BLUE
    GREEN = Palette.GREEN
    RED = Palette.RED
    YELLOW = Palette.YELLOW

    def get_colors(self) -> Color:
        return Color(
            colors_enabled=self.colors_enabled,
            palette=Palette(
                blue=DefaultTerm.BLUE,
                green=DefaultTerm.GREEN,
                red=DefaultTerm.RED,
                yellow=DefaultTerm.YELLOW
            )
        )



from sty import fg

from figcli.config.style.color import Color
from figcli.config.style.palette import Palette
from figcli.config.style.terms.term import Term


class iTerm(Term):
    TERM_PROGRAM = 'iTerm.app'
    BLUE = (52, 213, 235)
    GREEN = (51, 222, 136)
    RED = (240, 70, 87)
    YELLOW = (240, 240, 70)

    def get_colors(self) -> Color:
        return Color(
            colors_enabled=self.colors_enabled,
            palette=Palette(
                blue=iTerm.BLUE,
                green=iTerm.GREEN,
                red=iTerm.RED,
                yellow=iTerm.YELLOW
            )
        )



from figcli.commands.factory import Factory
from figcli.commands.help.login import Login
from figcli.commands.help.upgrade import Upgrade
from figcli.commands.help.version import Version
from figcli.commands.help_context import HelpContext
from figcli.commands.help.configure import Configure
from figcli.config import *
from figcli.svcs.setup import FiggySetup
from figcli.svcs.auth.session_manager import SessionManager
from figcli.utils.utils import Utils, CollectionUtils


class HelpFactory(Factory):
    def __init__(self, command: frozenset, context: HelpContext):
        self._command = command
        self._context = context
        self._options = context.options
        self._utils = Utils(False)
        self._setup: FiggySetup = FiggySetup()

    def instance(self):
        return self.get(self._command)

    def get(self, command: frozenset):
        if configure in self._options:
            return Configure(self._context, self._setup)
        elif version in self._options:
            return Version(self._context)
        elif command == login or command == sandbox:
            return Login(self._context, self._setup)
        elif upgrade in self._options:
            return Upgrade(self._context)
        else:
            self._utils.error_exit(f"{Utils.get_first(command)} is not a valid command. You must select from: "
                                   f"[{CollectionUtils.printable_set(help_commands)}]. Try using --help for more info.")



from configparser import ConfigParser
from enum import Enum
from typing import Type, Union, Optional, Callable, Any

from figcli.config import CONFIG_OVERRIDE_FILE_PATH, Config, EMPTY_CONFIG
from figcli.config.style.terminal_factory import TerminalFactory
from figcli.input import Input
from figcli.utils.utils import Utils
from threading import Lock
from os import path


class ConfigManager:
    """
    Offers functionality around storing/retrieving global user managed global configuration
    overrides. Operates on INI formatted configuration files.
    """

    def __init__(self, config_file: str):
        """
        :param config_file: /path/to/config_file
        """
        self.config_file = config_file
        self.config = None
        self._lock = Lock()
        self._load()

    def _load(self):
        self.config = ConfigParser()
        with open(self.config_file, 'r') as file:
            self.config.read_file(file)

    def set(self, key: Union[Type[Config.Section], Enum], val: Any) -> None:
        """
        Set the value in self.config_file for key: KEY
        Values are forced to strings if non-string values are passed
        :param key: Key to set in the file.
        :param val: Value to set.
        """
        # Update in case file changed on FS.
        with self._lock:
            self._load()

        if isinstance(val, bool):
            val = str(val).lower()
        else:
            val = str(val)

        self.config.set(key.__objclass__.NAME.value, key.value, val)

        with open(self.config_file, "w") as file:
            self.config.write(file)

    def get_or_prompt(self, key: Enum, get_it: Callable, colors_enabled=Utils.is_mac(), force_prompt=False) -> str:
        """
        Retrieves a value from the config_file based on the provided ENUM's value.
        If the value is unset, executes the get_it() method provided to retrieve the value, then returns it.

        If the user selects something _other_ than the default, overwrite the original value in the config file
        :param key: Enum representing the config value to fetch
        :param get_it: Method to execute if `config` is not in the configured config_file
        :param colors_enabled: Whether or not to enable colored output for the prompt if a config is found in config_file
        :return: String value from the config file, or the result of get_it()
        """
        c = TerminalFactory(colors_enabled).instance().get_colors()
        val = self.get_property(key)
        selection = None

        if val or force_prompt:
            if not force_prompt:
                print(f"\n\n{c.fg_bl}Default value found:{c.rs}")
                print(f"Key: {c.fg_gr}{key.value}{c.rs}")
                print(f"Value: {c.fg_gr}{val}{c.rs}")
                print(f"Values found in file: {c.fg_bl}{self.config_file}{c.rs}\n\n")

                selection = Input.y_n_input(f"Continue with `{val}`? ", default_yes=True)

            if not selection or force_prompt:
                original_val = val
                val = get_it()
                if val != original_val:
                    print(f"\n\nYou selected to overwrite the default. Updating default value:"
                          f"\nKey: {c.fg_gr}{key.value}{c.rs}"
                          f"\nFrom: {c.fg_yl}{original_val}{c.rs}"
                          f"\nTo: {c.fg_gr}{val}{c.rs}"
                          f"\nIn file {c.fg_bl}{self.config_file}{c.rs}")

                    self.set(key, val)

        else:
            val = get_it()

        return val

    def get_property(self, key: Union[Type[Config.Section], Enum]) -> Optional[str]:
        """
        Retrieves a property from the ini formatted config file and returns its value
        :param section: [section] in the ini config to find this parameter
        :param key: The key in the provided section to look up
        :param property: name of property to retrieve
        """
        try:
            if self.config.has_option(key.__objclass__.NAME.value, key.value):
                return self.config[key.__objclass__.NAME.value][key.value]
            else:
                return None
        except Exception as e:
            print(e)

    @staticmethod
    def figgy() -> "ConfigManager":
        """
        Get an instance of the figgy user provided configuration file
        """
        if not path.exists(CONFIG_OVERRIDE_FILE_PATH):
            with open(CONFIG_OVERRIDE_FILE_PATH, 'w') as figgy:
                figgy.write(EMPTY_CONFIG)

        return ConfigManager(CONFIG_OVERRIDE_FILE_PATH)



import json
import logging
import os
import platform
import re
import time
from collections import OrderedDict
from json.decoder import JSONDecodeError
from pathlib import Path
from sys import exit
from typing import Dict, List, Union

import botocore
import urllib3

from figcli.config import *
from figcli.config.style.color import Color
from figcli.config.style.terminal_factory import TerminalFactory

log = logging.getLogger(__name__)

BACKOFF = .25
MAX_RETRIES = 10


class Utils:

    def __init__(self, colors_enabled=False):
        self.c = TerminalFactory(colors_enabled).instance().get_colors()

    @staticmethod
    def retry(function):
        """
        Decorator that supports automatic retries if connectivity issues are detected with boto or urllib operations
        """

        def inner(self, *args, **kwargs):
            retries = 0
            while True:
                try:
                    return function(self, *args, **kwargs)
                except (botocore.exceptions.EndpointConnectionError, urllib3.exceptions.NewConnectionError) as e:
                    print(e)
                    if retries > MAX_RETRIES:
                        raise e

                    self._utils.notify("Network connectivity issues detected. Retrying with back off...")
                    retries += 1
                    time.sleep(retries * BACKOFF)

        return inner

    @staticmethod
    def trace(func):
        """
        Decorator that adds logging around function execution and function parameters.
        """

        def wrapper(*args, **kwargs):
            log.info(f"Entering function: {func.__name__} with args: {args}")
            start = time.time()
            result = func(*args, **kwargs)
            log.info(f"Exiting function: {func.__name__} and returning: {result}")
            log.info(f"Function complete after {round(time.time() - start, 2)} seconds.")
            return result

        return wrapper

    @staticmethod
    def millis_since_epoch():
        return int(time.time() * 1000)

    @staticmethod
    def get_os():
        return platform.system().lower()

    @staticmethod
    def not_windows():
        return platform != WINDOWS

    @staticmethod
    def is_linux():
        return platform.system() == LINUX

    @staticmethod
    def is_mac():
        return platform.system() == MAC

    @staticmethod
    def is_windows():
        return platform.system() == WINDOWS

    @staticmethod
    def find_figgy_json():
        for path in DEFAULT_FIGGY_JSON_PATHS:
            if Path(path).is_file():
                return path

        return 'figgy.json'

    @staticmethod
    def file_exists(path: str):
        return Path(path).is_file()

    @staticmethod
    def is_set_true(attr: frozenset, args) -> bool:
        attr_name = Utils.clean_attr_name(attr)

        if hasattr(args, attr_name):
            return args.__dict__.get(attr_name, False)
        else:
            return False

    @staticmethod
    def command_set(check_command: frozenset, args):
        command_name = args.command if hasattr(args, Utils.get_first(command)) else None

        return command_name == Utils.get_first(check_command)

    @staticmethod
    def attr_if_exists(attr: frozenset, args, default=None) -> Union[object, None]:
        attr_name = Utils.clean_attr_name(attr)
        return args.__dict__.get(attr_name, default)

    @staticmethod
    def attr_exists(attr: frozenset, args) -> bool:
        attr_name = Utils.clean_attr_name(attr)
        return args.__dict__.get(attr_name, None) is not None

    @staticmethod
    def clean_attr_name(attr: frozenset) -> str:
        return Utils.get_first(attr).replace('-', '_')

    @staticmethod
    def sanitize_session_name(name: str):
        return re.sub(r'\W+', '', name)[:15]


    @staticmethod
    def wipe_defaults():
        try:
            os.remove(DEFAULTS_FILE_CACHE_PATH)
        except OSError:
            pass

    @staticmethod
    def wipe_config_cache():
        try:
            os.remove(CONFIG_CACHE_FILE_PATH)
        except OSError:
            pass

    @staticmethod
    def wipe_vaults():
        for file in FIGGY_VAULT_FILES:
            try:
                os.remove(file)
            except OSError:
                pass

    def notify(self, message: str):
        print(f'{self.c.fg_bl}{message}{self.c.rs}')

    def merge_config_contents(self, a: Dict, b: Dict, a_path: str, b_path: str):
        for key in b:
            if isinstance(b[key], dict):
                if key in b and key in a:
                    dupes = b[key].keys() & a[key].keys()
                    self.validate(dupes == set(), f"Duplicate keys found between your configs. You may not have "
                                                  f"two instances of {key} with overlapping keys. Culprits: {dupes}")

                a_key = a[key] if key in a else {}
                b_key = b[key] if key in b else {}
                a[key] = {**a_key, **b_key}

            elif isinstance(b[key], list):
                a_key = a[key] if key in a else []
                b_key = b[key] if key in b else []
                a[key] = a_key + b_key
            else:
                self.error_exit(f"Unable to merge config values of type: {type(b[key])} for "
                                f"specified key {key} in both {a_path} and {b_path}")

        return a

    def get_repl_config(self, repl_config_path: str):
        with open(repl_config_path, "r") as file:
            contents = file.read()
            self.validate(contents != '', f"File provided at: {repl_config_path} "
                                          f"cannot be empty.")
            self.validate(self.is_json(contents), "File provided contains invalid json. Please remediate.")
            conf = json.loads(contents)
            self.validate(REPLICATION_KEY in conf, f"{REPLICATION_KEY} is missing from replication config: "
                                                   f"{repl_config_path}. This file is invalid.")
            return self.get_config_key_safe(REPLICATION_KEY, conf, default=[])

    def get_ci_config(self, ci_config_path: str) -> Dict:
        self.validate(ci_config_path.endswith('.json'),
                      "The figgy config file must end with the extension '.json'. A name of `figgy.json` is "
                      "recommended for most use cases..")

        self.validate(os.path.exists(ci_config_path), f"Path {ci_config_path} is invalid. That file does not exist.")

        # Read & Validate figgy.json
        try:
            with open(ci_config_path, "r") as file:
                base_matcher = re.search('^(.*[/]*).*.json$', ci_config_path)
                base_path = base_matcher.group(1)
                contents = file.read()
                self.validate(contents != '',
                              f"File provided at: {self.c.fg_rd}{ci_config_path}{self.c.rs} cannot be empty.")
                ci_config = json.loads(contents, encoding='utf-8')

                if IMPORTS_KEY in ci_config:
                    for import_val in ci_config[IMPORTS_KEY]:
                        import_path = f"{base_path}/{import_val}"
                        print(f"Loading imported config: {import_path}")
                        with open(import_path) as import_file:
                            contents = import_file.read()
                            imported_config = json.loads(contents)
                            ci_config = self.merge_config_contents(ci_config, imported_config, ci_config_path,
                                                                   import_path)

                namespace = self.get_namespace(ci_config)
                app_figs = self.get_config_key_safe(CONFIG_KEY, ci_config, default=[])
                dupes = self.find_dupes(app_figs)
                self.validate(not dupes, f"Your configuration has duplicate keys: {self.c.fg_rd}{dupes}{self.c.rs}")

                ns_app_params = self.standardize_parameters(namespace, app_figs)
                ci_config[CONFIG_KEY] = ns_app_params

                ns_shared_params = self.standardize_parameters(namespace,
                                                               self.get_config_key_safe(SHARED_KEY, ci_config,
                                                                                        default=[]))
                ci_config[SHARED_KEY] = ns_shared_params

                dupes = self.find_dupes(list(self.get_config_key_safe(REPLICATION_KEY, ci_config, default={}).values()))
                self.validate(not dupes, f"Your configuration has duplicate values in your replicated values "
                                         f"config: {self.c.fg_rd}{dupes}{self.c.rs}")

                self.validate(
                    len(self.get_config_key_safe(CONFIG_KEY, ci_config, default=[])) > 0
                    or OPTIONAL_NAMESPACE in ci_config, f"If you have no defined Names under: {CONFIG_KEY} you must "
                                                        f"specify an {OPTIONAL_NAMESPACE} parameter instead with a "
                                                        f"value of '/app/your-service-name/'")

                return ci_config

        except json.decoder.JSONDecodeError as e:
            print(f"{self.c.fg_rd}Error decoding json in figgy.json. Invalid JSON detected. "
                  f"Caught error: {e}{self.c.rs}")
            exit(1)
        except FileNotFoundError as e:
            print(f"{self.c.fg_rd}File at path {ci_config_path} does not exist or could not be read. "
                  f"Are you sure you provided a valid file path?{self.c.rs}")
            exit(1)

    def standardize_parameters(self, namespace: str, params: List[str]) -> List[str]:
        standardized = []
        for param in params:
            if not param.startswith(namespace):
                if param.startswith("/"):
                    standardized.append(f'{namespace}{param[1:]}')  # just in case they have an extra / by accident.
                else:
                    standardized.append(f'{namespace}{param}')
            else:
                standardized.append(param)

        return standardized

    def get_namespace(self, config: Dict):
        if OPTIONAL_NAMESPACE in config:
            namespace = config[OPTIONAL_NAMESPACE]
        else:
            self.validate(CONFIG_KEY in config, f"You must specify an {CONFIG_KEY} or {OPTIONAL_NAMESPACE} block, "
                                                f"or both, in your figgy.json file.")
            namespace = self.parse_namespace(self.get_first(set(config[CONFIG_KEY])))

        self.validate(namespace is not None, f"Invalid namespace provided, or unable to parse valid "
                                             f"namespace from your {CONFIG_KEY} block.")
        if not namespace.endswith('/'):
            namespace = namespace + '/'

        return namespace

    def get_service_name(self, config: Dict):
        namespace = self.get_namespace(config)
        return namespace.split('/')[2]

    @staticmethod
    def get_parameter_only(parameter_name: str):
        base_name = parameter_name
        """
        Takes /app/foo/a/full/path and returns a/full/path and always removes the namespace if it exists.
        :param parameter_name: name of a parameter, with or without the namespace.
        :return: parameter/name/path without any attached namespace.
        """
        try:
            get_param = re.compile(r"^/app/[A-Za-z0-9_-]+/(.*)")
            result = get_param.match(parameter_name)
            base_name = result.group(1)
        except (AttributeError, TypeError) as e:
            Utils.stc_error_exit(f"Unable to detect base name for parameter: {parameter_name}. {e}")

        return base_name

    @staticmethod
    def parse_namespace(app_key: str) -> str:
        ns = None
        try:
            get_ns = re.compile(r"^(/app/[A-Za-z0-9_-]+/).*")
            val = get_ns.match(app_key)
            ns = val.group(1)
        except (AttributeError, TypeError) as e:
            print(f"Unable to parse namespace from {app_key}. If your app_figs block values do not begin with "
                  f"the prefix /app/your-service-name , you must include the 'namespace' property in your figgy.json "
                  f"with value /app/your-service-name/")

        return ns

    def find_dupes(self, lst: List):
        return [x for n, x in enumerate(lst) if x in lst[:n]]

    def error_exit(self, error_msg: str):
        print(f"{self.c.fg_rd}ERROR: >> {error_msg}{self.c.rs}")
        exit(1)

    @staticmethod
    def stc_error_exit(error_msg: str):
        print(f"ERROR: >> {error_msg}")
        exit(1)

    @staticmethod
    def write_error(file_name: str, error_message: str):
        with open(f'{ERROR_LOG_DIR}/{file_name}', "w+") as log:
            log.write(error_message)

    @staticmethod
    def stc_validate(boolean: bool, error_msg: str):
        if not boolean:
            Utils().error_exit(error_msg)

    def validate(self, boolean: bool, error_msg: str):
        if not boolean:
            self.error_exit(error_msg)


    def is_valid_selection(self, selection: str, notify: bool):
        result = selection is not None and (selection.lower() == "y" or selection.lower() == "n")
        if notify and not result:
            msg = f"You must input a selection of 'Y' or 'N'"
            print(f"{self.c.fg_rd}{msg}{self.c.rs}")
        return result

    def is_valid_input(self, input: str, field_name: str, notify: bool):
        result = input is not None and input != ""
        if notify and not result:
            msg = f"ERROR: Your input of >> {input} << is not valid for {field_name}. " \
                  f"You cannot input an empty string or None"
            print(f"{self.c.fg_rd}{msg}{self.c.rs}")
        return result

    @staticmethod
    def stc_is_valid_input(input: str, field_name: str, notify: bool):
        result = input is not None and input != ""
        if notify and not result:
            msg = f"ERROR: Your input of >> {input} << is not valid for {field_name}. " \
                  f"You cannot input an empty string or None"
            print(f"{msg}")
        return result

    @staticmethod
    def format_config(config: Dict) -> OrderedDict:
        """
        Takes a formatted figgy.json file dictionary and converts it to an ordered dictionary. This makes it possible
        to write the file back out as a more readable and logically formatted file.
        """

        ordered_config = OrderedDict()

        if config.get(SERVICE_KEY):
            ordered_config[SERVICE_KEY] = config.get(SERVICE_KEY)

        if config.get(PLUGIN_KEY):
            ordered_config[PLUGIN_KEY] = config.get(PLUGIN_KEY)

        if config.get(OPTIONAL_NAMESPACE):
            ordered_config[OPTIONAL_NAMESPACE] = config.get(OPTIONAL_NAMESPACE)

        if config.get(IMPORTS_KEY):
            ordered_config[IMPORTS_KEY] = config.get(IMPORTS_KEY)

        ordered_config[CONFIG_KEY] = config.get(CONFIG_KEY, [])

        if config.get(REPL_FROM_KEY):
            ordered_config[REPL_FROM_KEY] = config.get(REPL_FROM_KEY)

        if config.get(REPLICATION_KEY):
            ordered_config[REPLICATION_KEY] = config.get(REPLICATION_KEY)

        if config.get(MERGE_KEY):
            ordered_config[MERGE_KEY] = config.get(MERGE_KEY)

        if config.get(SHARED_KEY):
            ordered_config[SHARED_KEY] = config.get(SHARED_KEY)

        return ordered_config

    @staticmethod
    def get_first(some_set: Union[set, frozenset]):
        if some_set:
            return set(some_set).pop()

        return None

    @staticmethod
    def str_too_long(value: str) -> bool:
        """
        Is this string too log to store in PS?
        Args:
            value: string to count

        Returns: bool
        """
        return len(value) > 4096

    @staticmethod
    def is_json(obj: str) -> bool:
        """
        Take a string and returns whether or not it is parseable as json.
        Args:
            obj: String to check

        Returns: True/False
        """
        try:
            json.loads(obj)
        except JSONDecodeError:
            return False

        # json.loads considers "true" or "false" valid json, but it's not valid in our case..
        if obj == "true" or obj == "false":
            return False

        # If it's a basic array, nope!
        if obj.startswith('[') and obj.endswith(']'):
            return False

        # json.loads considers strings that are numbers, valid json.
        try:
            int_obj = int(obj)
            if int_obj:
                return False
        # Do nothing if we catch an exception, That means this is json!
        except (NameError, ValueError):
            # json.loads considers strings that are numbers, valid json.
            try:
                fl_obj = float(obj)
                if fl_obj:
                    return False
            # Do nothing if we catch an exception, That means this is json!
            except (NameError, ValueError):
                return True

        return True

    @staticmethod
    def chunk_list(lst: List, chunk_size: int) -> List[List]:
        """Yield successive n-sized chunks from l."""
        for i in range(0, len(lst), chunk_size):
            yield lst[i:i + chunk_size]

    @staticmethod
    def parse_bool(value: Union[str, bool]) -> bool:
        if isinstance(value, bool):
            return bool(value)

        value = value.replace("'", '').replace('"', '').strip()
        if value.lower() == "true":
            return True
        elif value.lower() == "false":
            return False
        else:
            raise ValueError(f"Provided bool value of {value} is not a valid bool type.")

    @staticmethod
    def default_colors(enabled: bool = None) -> Color:
        if enabled is None:
            enabled = Utils.not_windows()

        return TerminalFactory(enabled).instance().get_colors()

    @staticmethod
    def to_env_var(variable_name: str):
        """
        converts aCasedVarable to A_CASED_VARIABLE case.
        """
        str1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', variable_name)
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', str1).upper()

    def get_config_key_safe(self, key: str, config: Dict, default=None):
        if key in config:
            return config[key]
        else:
            if default is None:
                return []
            else:
                return default


class InvalidSessionError(Exception):
    pass



from typing import Optional

from figcli.models.defaults.defaults import CLIDefaults
from figcli.models.run_env import RunEnv
from figcli.models.role import Role
from figcli.commands.command_context import CommandContext
from figcli.commands.figgy_context import FiggyContext
from figcli.utils.utils import Utils
from figcli.config import *


class ConfigContext(CommandContext):
    """
    All `config` resource commands require this context for general use. Config commands operate differently based on
    Role and RunEnv

    The context also contains optional parameter values passed in via the original invoked command vai the CLI
    """

    def __init__(self, run_env: RunEnv, role: Role, args, resource: frozenset, defaults: Optional[CLIDefaults]):
        super().__init__(run_env, resource, defaults=defaults)
        self.role = role  # type: Role
        self.args = args
        self.ci_config_path = Utils.attr_if_exists(config, args)
        self.from_path = Utils.attr_if_exists(from_path, args)
        self.out_file = Utils.attr_if_exists(out, args)
        self.prefix = Utils.attr_if_exists(prefix, args)
        self.service = Utils.attr_if_exists(service, args)

        # Flags like --prompt that are unset or set to true
        self.repl = Utils.is_set_true(replication_only, args)
        self.manual = Utils.is_set_true(manual, args)
        self.point_in_time = Utils.is_set_true(point_in_time, args)
        self.all_profiles = Utils.is_set_true(all_profiles, args)
        self.skip_upgrade = Utils.is_set_true(skip_upgrade, args)
        self.replication_only = Utils.is_set_true(replication_only, args)
        self.point_in_time = Utils.is_set_true(point_in_time, args)



from prompt_toolkit.shortcuts import prompt
from figcli.config import *
from figcli.data.dao.ssm import SsmDao
from figcli.data.dao.config import ConfigDao
from prompt_toolkit.completion import WordCompleter
from figcli.commands.types.config import ConfigCommand
from figcli.commands.config_context import ConfigContext
from figcli.config import *
from figcli.svcs.observability.anonymous_usage_tracker import AnonymousUsageTracker
from figcli.svcs.observability.version_tracker import VersionTracker
from figcli.utils.utils import *


class Dump(ConfigCommand):
    """
    Allows users to dump PS K/V hierarchy into JSON, straight to the terminal, or to a file itself.
    """

    def __init__(self, ssm_init: SsmDao, config_completer_init: WordCompleter, colors_enabled: bool,
                 context: ConfigContext):
        super().__init__(audit, colors_enabled, context)
        self._ssm = ssm_init
        self._config_completer = config_completer_init
        self._utils = Utils(colors_enabled)
        self._output_file = context.out_file
        self._prefix = context.prefix
        self.example = f"{self.c.fg_bl}{CLI_NAME} config {self.type} --env dev{self.c.rs} --prefix /app/demo-time " \
            f"--out /tmp/out_file.json"

    def _dump(self):
        usr_prefix = self._prefix  # type: str
        notify = False  # type: bool

        while not self._utils.is_valid_input(usr_prefix, 'Prefix', notify) \
                and not self._prefix:
            usr_prefix = prompt(f"Please input a Prefix to dump from: ", completer=self._config_completer)

        all_names = list(map(lambda x: x['Name'], self._ssm.get_all_parameters([usr_prefix])))  # type: List[str]
        all_vals = self._ssm.get_parameter_values(all_names, decrypt=False)  # type: List[Dict]
        json_output = {}
        for val in all_vals:
            json_output[val['Name']] = val['Value']

        if self._output_file:
            with open(self._output_file, "w") as file:
                json.dump(json_output, file, indent=4, sort_keys=True)
        else:
            print(json.dumps(json_output, indent=4, sort_keys=True))


    @VersionTracker.notify_user
    @AnonymousUsageTracker.track_command_usage
    def execute(self):
        self._dump()



import pexpect
from figcli.test.cli.config import *
from figcli.test.cli.figgy import FiggyTest
from figcli.config import *
from figcli.utils.utils import *


class DevGet(FiggyTest):

    def __init__(self, extra_args=""):
        super().__init__(pexpect.spawn(f'{CLI_NAME} config {Utils.get_first(get)} --env {DEFAULT_ENV} '
                                       f'--skip-upgrade {extra_args}',
                                       timeout=10, encoding='utf-8'), extra_args=extra_args)
        self._child.delayafterread = .01
        self._child.delaybeforesend = .5

    def run(self):
        self.step(f"Testing GET for {param_1}")
        self.get(param_1, param_1_val, get_more=False)

    def get(self, key, value, get_more=False, expect_missing=False, no_decrypt=False):
        self.expect('.*PS Name*')
        self.sendline(key)
        if expect_missing:
            self.expect(f'.*Invalid PS Name specified..*')
            print("Missing parameter validated.")
        elif no_decrypt:
            self.expect(f'.*do not have access.*')
            print("Lack of decrypt permissions validated.")
        else:
            self.expect(f'.*{value}.*Get another.*')
            print(f"Expected value of {value} validated.")
        if get_more:
            print("Getting another")
            self.sendline('y')
        else:
            print("Not getting another")
            self.sendline('n')



import click
from botocore.exceptions import ClientError
from prompt_toolkit import prompt
from prompt_toolkit.completion import WordCompleter

from figcli.commands.config.get import Get
from figcli.commands.config_context import ConfigContext
from figcli.commands.types.config import ConfigCommand
from figcli.config.style.style import FIGGY_STYLE
from figcli.data.dao.ssm import SsmDao
from figcli.svcs.observability.anonymous_usage_tracker import AnonymousUsageTracker
from figcli.svcs.observability.version_tracker import VersionTracker
from figcli.utils.utils import *
from figcli.views.rbac_limited_config import RBACLimitedConfigView


class List(ConfigCommand):

    def __init__(self, config_view: RBACLimitedConfigView, config_completer_init: WordCompleter,
                 colors_enabled: bool, config_context: ConfigContext, get: Get):
        super().__init__(list_com, colors_enabled, config_context)
        self._view = config_view
        self._config_completer = config_completer_init
        self._get = get
        self._utils = Utils(colors_enabled)

    def _list_params(self):
        """
        Prompts user for a namespace of parameters to query. Once queried, user may select from the outputted chart
        of found parameters matching that namespace. This will then provide the user the selected value.
        """

        # Prompts for this file
        select_message = [
            (f'class:{self.c.bl}', '`[1-999]` '),
            ('class:', 'to select a value. '),
            (f'class:{self.c.bl}', '`continue` '),
            ('class:', 'to query more. Anything else quits. \n\nSelection: ')
        ]

        # Add all keys
        list_another = True

        while list_another:
            print()
            namespace = prompt(f"Please input a namespace prefix to query: ", completer=self._config_completer)
            if not self._utils.is_valid_input(namespace, "namespace", notify=False):
                continue

            print(f"Passing in prefix of :{namespace}")
            parameters = self._view.get_config_names(prefix=namespace)
            names = [['Selector', 'Name']]
            count = 1
            for param in parameters:
                names.append([f'{self.c.fg_rd}{count}{self.c.rs}', param])
                count = count + 1

            # Pretty print out with columns for selection.
            data = '\n'.join(['\t'.join([str(cell) for cell in row]) for row in names])
            keep_getting = True

            while keep_getting:
                pagination_threshold = 30
                if count > pagination_threshold:
                    pagination_data = f"Find your number, then use \"q\" to quit. Then input your number.\n\n {data}"
                    click.echo_via_pager(pagination_data)
                else:
                    pagination_data = data
                    click.echo(pagination_data)

                selection = prompt(select_message, style=FIGGY_STYLE)
                if re.match('[0-9]+', selection) is not None:
                    if int(selection) > count - 1:
                        print(f"Invalid selection. try again")
                        continue
                    print(f"Name: {names[int(selection)][1]}")
                    value, desc = self._get.get_val_and_desc(names[int(selection)][1])
                    desc = desc if desc else DESC_MISSING_TEXT

                    if value is not None:
                        print(f"Value: {value}")
                        print(f"{self.c.fg_gr}Description: {self.c.rs}{desc}")

                    hold = input(f"\nPress ENTER when you want to look up another value.")
                    click.clear()
                else:
                    list_another = selection.lower() == "continue"
                    keep_getting = False

    @VersionTracker.notify_user
    @AnonymousUsageTracker.track_command_usage
    def execute(self):
        self._list_params()



import sys

import pexpect
from figcli.test.cli.config import *
from figcli.test.cli.figgy import FiggyTest
from figcli.test.cli.dev.put import DevPut
from figcli.test.cli.dev.delete import DevDelete
from figcli.test.cli.dev.get import DevGet
from figcli.test.cli.dev.audit import DevAudit
from figcli.config import *
from figcli.utils.utils import *
import time
import uuid

RESTORE_PROPAGATION_TIME = 120


class DevRestore(FiggyTest):

    def __init__(self, extra_args=""):
        super().__init__(None, extra_args=extra_args)
        self._guuid = uuid.uuid4().hex
        print(f"Guuid generated: {self._guuid}")

    def run(self):
        minimum, maximum = 1, 3
        first_val = 'FIRST_VAL'
        second_val = 'SECOND_VAL'
        self.step("Adding new configs. To Restore")
        self._setup(minimum, maximum, value_override=first_val)
        print("Waiting for propagation...")
        time.sleep(RESTORE_PROPAGATION_TIME)
        print("Validating")
        self._audit(minimum, maximum)
        restore_breakpoint_1 = int(time.time() * 1000)
        print(f"First restore breakpoint: {restore_breakpoint_1} - would expect val of {first_val}")

        time.sleep(15)
        self._setup(minimum, maximum, value_override=second_val)
        print("Waiting for propagation...")
        time.sleep(RESTORE_PROPAGATION_TIME)
        print("Validating")
        self._audit(minimum, maximum)
        restore_breakpoint_2 = int(time.time() * 1000)
        print(f"Second restore breakpoint: {restore_breakpoint_1} - would expect val of {second_val}")

        time.sleep(15)

        restore_prefix = f'{param_test_prefix}{self._guuid}/'
        self.step(f"Attempting restore to time: {restore_breakpoint_1} with prefix: {restore_prefix}")
        child = pexpect.spawn(f'{CLI_NAME} config {Utils.get_first(restore)} --env {DEFAULT_ENV} --skip-upgrade'
                              f' --point-in-time {self.extra_args}', timeout=5, encoding='utf-8')
        child.expect('.*Which.*recursively restore.*')
        child.sendline(restore_prefix)
        child.expect('.*Are you sure you want.*')
        child.sendline('y')
        child.expect('.*Seconds.*restore.*')
        child.sendline(f'{restore_breakpoint_1}')
        child.expect('.*Are you sure.*')
        child.sendline('y')
        print("Checking restore output...\r\n\r\n")
        child.expect(f'.*Value.*{param_test_prefix}{self._guuid}/test_param.*current.*Skipping.*')

        self.step("Validating values were rolled back...")
        time.sleep(5)

        get = DevGet(extra_args=self.extra_args)
        get.get(f'{param_test_prefix}{self._guuid}/test_param', 'NOT_CHANGED_VAL', get_more=True)
        for i in range(minimum, maximum):
            get.get(f'{param_test_prefix}{self._guuid}/test_param-{i}', first_val, get_more=i < maximum - 1)

        print("Values were rolled back successfully")
        time.sleep(30)
        print("Testing restore to second restore point")

        restore_prefix = f'{param_test_prefix}{self._guuid}/'
        self.step(f"Attempting restore to time: {restore_breakpoint_2} with prefix: {restore_prefix}")
        child = pexpect.spawn(f'{CLI_NAME} config {Utils.get_first(restore)} --env {DEFAULT_ENV} --skip-upgrade'
                              f' --point-in-time {self.extra_args}', timeout=5, encoding='utf-8')
        child.expect('.*Which.*recursively restore.*')
        child.sendline(restore_prefix)
        child.expect('.*Are you sure you want.*')
        child.sendline('y')
        child.expect('.*Seconds.*restore.*')
        child.sendline(f'{restore_breakpoint_2}')
        child.expect('.*Are you sure.*')
        child.sendline('y')
        print("Checking restore output...\r\n\r\n")
        child.expect(f'.*Restoring.*{param_test_prefix}{self._guuid}/test_param-2.*{second_val}.*')

        self.step("Validating values were rolled back...")
        time.sleep(5)

        get = DevGet(extra_args=self.extra_args)
        get.get(f'{param_test_prefix}{self._guuid}/test_param', 'NOT_CHANGED_VAL', get_more=True)
        for i in range(minimum, maximum):
            get.get(f'{param_test_prefix}{self._guuid}/test_param-{i}', second_val, get_more=i < maximum - 1)

        print("Values were rolled forward successfully. Cleaning up...")

        delete = DevDelete(extra_args=self.extra_args)
        delete.delete(f'{param_test_prefix}{self._guuid}/test_param', delete_another=True)
        for i in range(minimum, maximum):
            delete.delete(f'{param_test_prefix}{self._guuid}/test_param-{i}', delete_another=i < maximum - 1)

    def _setup(self, min: int, max: int, value_override: str = None):
        put = DevPut(extra_args=self.extra_args)
        put.add(f'{param_test_prefix}{self._guuid}/test_param', 'NOT_CHANGED_VAL', param_1_desc, add_more=True)
        for i in range(min, max):
            value = value_override if value_override else DELETE_ME_VALUE
            put.add_another(f'{param_test_prefix}{self._guuid}/test_param-{i}', value, f'{param_1_desc}-{i}',
                            add_more=i < max - 1)

        get = DevGet(extra_args=self.extra_args)
        get.get(f'{param_test_prefix}{self._guuid}/test_param', 'NOT_CHANGED_VAL', get_more=True)
        for i in range(min, max):
            value = value_override if value_override else DELETE_ME_VALUE
            get.get(f'{param_test_prefix}{self._guuid}/test_param-{i}', value, get_more=i < max - 1)

    def _audit(self, min: int, max: int):
        audit = DevAudit(extra_args=self.extra_args)
        audit.audit(f'{param_test_prefix}{self._guuid}/test_param', audit_another=True)

        for i in range(min, max):
            audit.audit(f'{param_test_prefix}{self._guuid}/test_param-{i}', audit_another=i < max - 1)



from figcli.models.run_env import RunEnv
from figcli.config import *
import getpass
from typing import Dict, List
from figcli.utils.utils import *


class ReplicationType:
    """
    Represents the various types of replication configs that are valid for storing into our dynamodb
    `service-config-replication` table
    """

    def __init__(self, type: str):
        assert type in repl_types, f"Type: {type} must be one of: {repl_types}"
        self.type = type


class ReplicationConfig:
    """
    This model is used for storing / retrieving data from the `service-config-replication` table.
    """

    def __init__(self, destination: str, run_env: RunEnv, namespace: str, source: str, type: ReplicationType,
                 user: str = None):
        self.destination = destination
        self.run_env = run_env.env
        self.namespace = namespace
        self.source = source
        self.type = type.type
        self.user = user

        if user is None:
            self.user = getpass.getuser()

        self.props = {
            REPL_SOURCE_ATTR_NAME: self.source,
            REPL_NAMESPACE_ATTR_NAME: self.namespace,
            REPL_TYPE_ATTR_NAME: self.type,
            REPL_USER_ATTR_NAME: self.user
        }

    @staticmethod
    def from_dict(conf: Dict, type: ReplicationType, run_env: RunEnv,
                  namespace: str = None, user: str = None) -> List:
        """
        Dict must be of format - Key (source) -> Value (destination)
        Args:
            conf: Key (repl source) -> Value (repl dest) dictionary
            type: required - Type of replication config (merge / app)
            run_env: RunEnvironment, optional
            namespace: Dest App Namespace, optional
            user: User who is creating this REPL Conf, also optional
        Returns:
            List[ReplicationConfig] - List of hydrated replication config objects based on the parameters.
        """
        cfgs = []
        for key in conf:
            if namespace is None:
                namespace = Utils.parse_namespace(conf[key])
            if user is None:
                user = getpass.getuser()
            cfgs.append(ReplicationConfig(destination=conf[key], source=key, type=type,
                                          run_env=run_env, namespace=namespace, user=user))
        return cfgs

    def __str__(self):
        return f"{self.__dict__}"

    def __hash__(self):
        return hash(f"{self.destination}{self.source}{self.type}")

    def __eq__(self, other):
        if isinstance(other, ReplicationConfig):
            return self.destination == other.destination and self.source == other.source and self.type == other.type
        return False



from typing import List
import boto3
import botocore

from figcli.models.assumable_role import AssumableRole
from figcli.models.defaults.defaults import CLIDefaults
from figcli.svcs.auth.provider.session_provider import SessionProvider
from figcli.utils.utils import Utils

class ProfileSessionProvider(SessionProvider):
    """
    This provider is not recommended; however, I recognize it is useful for 2 main use cases.

    1) Companies with a single AWS account.
    2) Automating CICD pipelines.

    This provider always returns a session from the targeted profile.
    """

    def __init__(self, defaults: CLIDefaults):
        super().__init__(defaults)

    def get_session(self, assumable_role: AssumableRole, prompt: bool, exit_on_fail=True) -> boto3.Session:

        try:
            return boto3.Session(profile_name=assumable_role.profile)
        except botocore.exceptions.ProfileNotFound:
            Utils.stc_error_exit(f"The provided profile override of {assumable_role.profile} is invalid. Are you sure "
                                 f"this profile is set in your ~/.aws/credentials and ~/.aws/config files?")

    def cleanup_session_cache(self):
        pass



import pexpect
from figcli.test.cli.config import *
from figcli.test.cli.figgy import FiggyTest
from figcli.utils.utils import *
import os


class ConfigureGoogle(FiggyTest):
    def __init__(self, role_type: str):
        self._role_type = role_type
        super().__init__(None)

        self._child = pexpect.spawn(f'{CLI_NAME} --{Utils.get_first(configure)} --skip-upgrade {self.extra_args}',
                              encoding='utf-8', timeout=5)

    def run(self):
        self.step(f"Testing `{CLI_NAME} --{Utils.get_first(configure)}`")
        user_name = os.environ.get(GOOGLE_SSO_USER)
        password = os.environ.get(GOOGLE_SSO_PASSWORD)
        idp_id = os.environ.get(GOOGLE_IDP_ID)
        sp_id = os.environ.get(GOOGLE_SP_ID)

        self._child.expect('.*select.*GOOGLE.*')
        self._child.sendline('GOOGLE')
        self._child.expect('.*GOOGLE username.*')
        self._child.sendline(user_name)
        self._child.expect('.*GOOGLE password.*')
        self._child.sendline(password)
        self._child.expect('.*mfa_enabled.*')
        self._child.sendline('y')
        self._child.expect('.*Google Account.*')
        self._child.sendline(idp_id)
        self._child.expect('.*Please input.*Provider ID.*')
        self._child.sendline(sp_id)
        self._child.expect('.*overwrite.*REDACTED.*')
        self._child.sendline(self._role_type)
        self._child.expect('.*Options.*dev.*')
        self._child.sendline('dev')
        self._child.expect('.*region.*')
        self._child.sendline('y')
        self._child.expect('.*colors.*')
        self._child.sendline('n')
        self._child.expect('.*weirdness.*')
        self._child.sendline('n')
        self._child.expect('.*anonymous_metrics_enabled.*')
        self._child.sendline('n')
        self._child.expect('.*anonymou.*')
        self._child.sendline('n')
        self._child.expect('.*overwrite.*')
        self._child.expect('.*Setup successful.*')
        self.step(f"GOOGLE with role: {self._role_type} configured successfully")



from figcli.config import *
from figcli.utils.utils import *


class AWSCLIUtils:
    """
    Utility methods for interacting with AWSCLI resources, such as the ~/.aws/credentials and ~/.aws/config files
    """

    @staticmethod
    def write_credentials(access_key: str, secret_key: str, token: str, region: str,
                          profile_name: str = 'default', color: Color = Color(False)) -> None:
        """
        Overwrite credentials stored in the [default] profile in both ~/.aws/config and ~/.aws/credentials file
        with the provided temporary credentials. This method also CREATES these files if they do not already exist.
        """

        credentials_contents, config_contents = [], []
        default_keys = [f"[{profile_name}]\n",
                        f"aws_access_key_id = {access_key}\n",
                        f"aws_secret_access_key = {secret_key}\n",
                        f"aws_session_token = {token}\n\n"]

        default_config = [f"[profile {profile_name}]\n",
                          "output = json\n",
                          f"region = {region}\n"]

        try:
            with open(AWS_CREDENTIALS_FILE_PATH, "r") as aws_creds:
                credentials_contents = aws_creds.readlines()

        except FileNotFoundError:
            print(f"File missing: {AWS_CREDENTIALS_FILE_PATH} - will be created.")
        finally:
            if not credentials_contents:
                contents = default_keys
            else:
                contents = AWSCLIUtils._replace_profile(f"[{profile_name}]\n", credentials_contents,
                                                        replace_with=default_keys)

            os.makedirs(os.path.dirname(AWS_CREDENTIALS_FILE_PATH), exist_ok=True)
            with open(AWS_CREDENTIALS_FILE_PATH, "w+") as creds:
                creds.writelines(contents)

            print(f"\n{color.fg_gr}Successfully updated: {AWS_CREDENTIALS_FILE_PATH}{color.rs}")

        try:
            with open(AWS_CONFIG_FILE_PATH, "r") as aws_config:
                config_contents = aws_config.readlines()
        except FileNotFoundError:
            print(f"File missing: {AWS_CONFIG_FILE_PATH} - will be created.")
        finally:
            if not config_contents:
                contents = default_config
            else:
                contents = AWSCLIUtils._replace_profile(f"[profile {profile_name}]\n", config_contents,
                                                        replace_with=default_config)

            os.makedirs(os.path.dirname(AWS_CONFIG_FILE_PATH), exist_ok=True)
            with open(AWS_CONFIG_FILE_PATH, "w+") as creds:
                creds.writelines(contents)

            print(f"{color.fg_gr}Successfully updated: {AWS_CONFIG_FILE_PATH}{color.rs}")

    @staticmethod
    def _replace_profile(profile_header: str, contents: List[str], replace_with: List[str]) -> List[str]:
        """
        Replaces a profile in the provided file contents of an ~/.aws/config or ~/.aws/credentials file
        Args:
            profile_header: e.g. [default] or [profile default] - the AWS config file header to replace
            contents: List[str] The contents of the file
            replace_with: The new profile data to substitute for the previous profile
        """
        if profile_header in contents:
            default_index = contents.index(profile_header)
            del contents[default_index]
            while "[" not in contents[default_index]:
                del contents[default_index]
                if default_index > len(contents) - 1:
                    break

        contents = replace_with + ['\n'] + contents
        return contents



import time
import logging
from dataclasses import dataclass, field
import requests
from typing import List, Dict

from figcli.models.defaults.defaults import CLIDefaults
from figcli.svcs.cache_manager import CacheManager
from figcli.svcs.setup import FiggySetup
from figcli.utils.utils import Utils
from concurrent.futures import ThreadPoolExecutor
from figcli.config import *

log = logging.getLogger(__name__)


@dataclass
class FiggyMetrics:
    COUNT_KEY = 'count'
    user_id: str
    metrics: Dict[str, Dict] = field(default_factory=dict)
    last_report: int = Utils.millis_since_epoch()

    def increment_count(self, command: str) -> None:
        metric = self.metrics.get(command, {})
        metric[FiggyMetrics.COUNT_KEY] = metric.get(FiggyMetrics.COUNT_KEY, 0) + 1
        self.metrics[command] = metric


class AnonymousUsageTracker:
    """
    We want to track the usage counts of various commands and the version of Figgy people are currently using.
    This data will be valuable for informing future decisions & when considering upgrade paths or potentially
    introducing breaking changes and their impacts.
    """
    _CACHE_NAME = 'usage-metrics'
    _METRICS_KEY = 'metrics'
    _USER_KEY = 'user_id'
    _VERSION_KEY = 'version'
    REPORT_FREQUENCY = 60  # Report daily
    # REPORT_FREQUENCY = 1000 * 60 * 60 * 24  # Report daily

    @staticmethod
    def report_usage(metrics: FiggyMetrics):
        metrics_json = {AnonymousUsageTracker._METRICS_KEY: {}, AnonymousUsageTracker._USER_KEY: metrics.user_id}
        for key, val in metrics.metrics.items():
            metrics_json[AnonymousUsageTracker._METRICS_KEY][key] = val.get(FiggyMetrics.COUNT_KEY, 0)

        metrics_json[AnonymousUsageTracker._VERSION_KEY] = VERSION

        requests.post(url=FIGGY_LOG_METRICS_URL, json=metrics_json)

    @staticmethod
    def track_command_usage(function):
        """
        Tracks user command usage locally. This will be intermittently reported in aggregate.
        """

        def inner(self, *args, **kwargs):
            command = getattr(self, 'type', None)
            if command:
                command = Utils.get_first(command)
                cache = CacheManager(AnonymousUsageTracker._CACHE_NAME)

                if hasattr(self, 'context') and hasattr(self.context, 'defaults') and self.context.defaults is not None:
                    if isinstance(self.context.defaults, CLIDefaults):
                        user_id = self.context.defaults.user_id
                    else:
                        user_id = "EmptyDefaults"
                else:
                    user_id = "NoOne"

                last_write, metrics = cache.get(AnonymousUsageTracker._METRICS_KEY,
                                                default=FiggyMetrics(user_id=user_id))

                metrics.increment_count(command)
                if Utils.millis_since_epoch() - metrics.last_report > AnonymousUsageTracker.REPORT_FREQUENCY:
                    defaults = FiggySetup.stc_get_defaults(skip=True)
                    if defaults and defaults.usage_tracking:
                        # Ship it async. If it don't worky, oh well :shruggie:
                        with ThreadPoolExecutor(max_workers=1) as pool:
                            pool.submit(AnonymousUsageTracker.report_usage, metrics)
                            cache.write(AnonymousUsageTracker._METRICS_KEY, FiggyMetrics(user_id=user_id))
                            return function(self, *args, **kwargs)
                else:
                    cache.write(AnonymousUsageTracker._METRICS_KEY, metrics)

            return function(self, *args, **kwargs)

        return inner



import sys

import pexpect
from figcli.test.cli.config import *
from figcli.test.cli.dev.get import DevGet
from figcli.test.cli.dev.delete import DevDelete
from figcli.test.cli.dev.put import DevPut

from figcli.test.cli.figgy import FiggyTest
from figcli.config import *
from figcli.utils.utils import *
import uuid
import time

AUDIT_PROPAGATION_TIME = 60

class DevAudit(FiggyTest):
    def __init__(self, extra_args=""):
        super().__init__(None, extra_args=extra_args)

    def run(self):
        put = DevPut(extra_args=self.extra_args)
        guuid = uuid.uuid4().hex
        key = f"{param_test_prefix}{guuid}"
        put.add(key, DELETE_ME_VALUE, 'desc', add_more=False)
        get = DevGet(extra_args=self.extra_args)
        get.get(key, DELETE_ME_VALUE, get_more=False)
        self.step(f"Sleeping {AUDIT_PROPAGATION_TIME} to allow for lambda -> dynamo audit log insert.")
        time.sleep(AUDIT_PROPAGATION_TIME)
        self.step(f"Looking up audit log for: {key}. If this fails, the lambda could be broken. ")
        self.audit(key)
        delete = DevDelete(extra_args=self.extra_args)
        delete.delete(key)

        new_uuid = uuid.uuid4().hex
        self.step("Testing searching for non-existent audit log.")
        self.audit(f'/doesnt/exist/{new_uuid}', expect_results=False)

    def audit(self, name, audit_another=False, expect_results=True):
        child = pexpect.spawn(f'{CLI_NAME} config {Utils.get_first(audit)} --env {DEFAULT_ENV} --skip-upgrade'
                              f' {self.extra_args}', encoding='utf-8', timeout=10)
        child.logfile = sys.stdout

        self.step(f"Auditing: {name}")
        child.expect('.*Please.*input.*')
        child.sendline(name)
        if expect_results:
            child.expect(f'.*Found.*Parameter:.*{name}.*Audit another.*')
            print("Audit log found.")
        else:
            child.expect(f'.*No results found for.*{name}.*')
            print("No audit log found.")

        if audit_another:
            child.sendline('y')
        else:
            child.sendline('n')




class RunEnv:

    def __init__(self, env: str, account_id: str = None):
        self.env = env
        self.account_id = account_id

    def __str__(self):
        return self.env

    def __eq__(self, obj):
        if isinstance(obj, RunEnv):
            return obj.env == self.env

        return False

    def __hash__(self):
        return hash(self.env)



from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass
from threading import Lock
from typing import Tuple, List, Union

import boto3
from botocore.response import StreamingBody
from tqdm import tqdm

from figcli.utils.utils import *

log = logging.getLogger(__name__)


@dataclass(repr=True)
class S3MatchResult:
    def __init__(self, s3_key: str,
                 matched_text: Union[str, None],
                 matched_context: Union[str, None],
                 skipped: bool = False):
        self.s3_key = s3_key
        self.matched_text = matched_text
        self._matched_context = matched_context
        self.skipped = skipped

    @property
    def matched_context(self):
        return self._matched_context

    @matched_context.setter
    def matched_context(self, text: str):
        self._matched_context = text if text else "N/A"

    def has_match(self) -> bool:
        return self.matched_text is not None

    def __hash__(self):
        return hash(self.s3_key)

    def __eq__(self, o):
        if isinstance(o, S3MatchResult):
            return o.s3_key == self.s3_key


class S3Service:
    _MAX_POOL_SIZE = 300
    """
    Provides functionality around S3
    """
    _SEARCHABLE_CONTENT_TYPES = ['application/json', 'application/xml', 'text/html', 'text/css', 'text/javascript',
                                 'text/csv']
    _SEARCHABLE_SUFFIXES = ['.json', '.txt', '.csv', '.xml', '.log']

    def __init__(self, env_session: boto3.Session):
        self._s3 = env_session.client('s3', config=botocore.client.Config(max_pool_connections=S3Service._MAX_POOL_SIZE))
        self._s3_resource = env_session.resource('s3')
        self.__progress_total = 0
        self.lock = Lock()

    @property
    def progress_total(self):
        with self.lock:
            return self.__progress_total

    def get_buckets(self) -> List[str]:
        """
        :return: List[str] = A list of bucket names for the account linked to the s3 client.
        """
        response = self._s3.list_buckets()

        assert 'Buckets' in response, "No buckets found in S3 response."

        buckets = response.get('Buckets', [])

        # Returns bucket names and strips out missing bucket names if any (would be probably rare or impossible)
        return [bucket.get('Name') for bucket in buckets if bucket.get('Name')]

    def get_folders(self, bucket_name: str) -> List[str]:
        """
        Return top level folders in the given S3 bucket.
        :param bucket_name: S3 bucket to search
        :return: List[str] - A list of folders that live in the top level of a S3 bucket.
        """

        paginator = self._s3.get_paginator('list_objects')
        result = paginator.paginate(Bucket=bucket_name, Delimiter='/')
        folders = []
        for prefix in result.search('CommonPrefixes'):
            folders.append(prefix.get('Prefix'))

        return folders

    def get_folders_recursive(self, bucket_name: str, folder_prefix: str = None, max_recursion_depth: int = 500,
                              recursion_depth: int = 0, max_folders_per_recurse: int = 500) -> Tuple[List[str], int]:
        """
        Find all subfolders in an S3 bucket by recursively iterating through each folder.
        :param bucket_name: Bucket to search
        :param folder_prefix: Used for recursive searches
        :param max_recursion_depth: Tunable parameter that allows control over recursion depth.
        :param recursion_depth: DO NOT PASS THIS PARAM. Keeps track of recursion depth to prevent bucket traversal
               going on FOREVER
        :param max_folders_per_recurse: Define a maximum # of folders to find per recursion depth. This prevents
        a single directory with a billion subdirs from holding up the recursion indefinitely. Set to 0 for infinite.

        :return: List[str] - a List of folders found in the S3 bucket.
        """

        recursion_depth = recursion_depth + 1
        paginator = self._s3.get_paginator('list_objects')
        new_folders, all_folders = [], []
        if not folder_prefix:
            result = paginator.paginate(Bucket=bucket_name, Delimiter='/')
        else:
            result = paginator.paginate(Bucket=bucket_name, Prefix=folder_prefix, Delimiter='/')

        added_count = 0
        with ThreadPoolExecutor(max_workers=100) as pool:
            for prefix in result.search('CommonPrefixes'):
                if prefix:  # Prefix can sometimes be None
                    new_folders.append(prefix.get('Prefix'))
                    log.info(f"Adding folder: {prefix} - #{added_count}")
                    added_count += 1

            all_futures = []
            all_folders = new_folders.copy()
            if recursion_depth < max_recursion_depth:
                count = 0
                for folder in new_folders:
                    count = count + 1
                    if count < max_folders_per_recurse or max_folders_per_recurse == 0:
                        future = pool.submit(self.get_folders_recursive, bucket_name=bucket_name,
                                             folder_prefix=folder, max_recursion_depth=max_recursion_depth,
                                             recursion_depth=recursion_depth)
                        all_futures.append(future)
                    else:
                        log.info("Hit max recursion limit, skipping more recursion to search for more folders.")
                        break

                for future in as_completed(all_futures):
                    next_level, traversed_depth = future.result()
                    all_folders = all_folders + next_level
                    recursion_depth = traversed_depth if traversed_depth > recursion_depth else recursion_depth

        log.info(f"Returning from recursion into: {folder_prefix}")
        return sorted(list(set(all_folders))), recursion_depth

    def __list_progress_bar(self, file_count, enable_output: bool = True):
        for i in tqdm(range(0, file_count), disable=not enable_output, desc="Aggregating files...    ",
                      unit="folders aggregated."):
            while self.__progress_total < i:
                time.sleep(.1)

    def list_files(self, bucket_name: str, folder_prefix: str, folder_filter: str = None, enable_output=False,
                   max_recursion_depth: int = 50):
        """
        Return a list of all files in a bucket/folder prefix. The file names are the full path, including the prefix.
            Standard list with pagination will be very slow on large folders. This approach will return faster across
            large data sets by breaking up lists by each folder and running concurrently.
        :param bucket_name: bucket_name to query
        :param folder_prefix: some/path/to/query/ - folder heirarchy to query
        :param folder_filter: Optional[str] - only list files in folders that match this filter string. Wildcards (*)
                accepted.
        :param enable_output: Turns out printing details of search during search process to stdout.
        :param max_recursion_depth: Depth of folders to traverse while listing files.
        :return: List[str] List of file paths in S3 that match.
        """

        start_time  = time.time()
        paginator = self._s3.get_paginator('list_objects')
        all_files = []

        subfolders, depth_reached = self.get_folders_recursive(bucket_name, folder_prefix,
                                                               max_recursion_depth=max_recursion_depth)

        log.info(f'Found {len(subfolders)} searchable subfolders.')
        subfolders.append(folder_prefix)
        enable_output and print(f"Found {len(subfolders)} folders that could be searched.")

        # Filter our all non-matching folder_filter
        if folder_filter:
            subfolders = [x for x in subfolders if re.match(f'.*({folder_filter.replace("*", ".*")}).*', x)]
            enable_output and print(f"Searching {len(subfolders)} folders after filtering applied."
                                    f" Aggregating searchable file list.")

        futures = set()
        with ThreadPoolExecutor(max_workers=100) as pool:
            pool.submit(self.__list_progress_bar, len(subfolders), enable_output)

            for folder in subfolders:
                futures.add(pool.submit(paginator.paginate, Bucket=bucket_name, Delimiter='/', Prefix=folder))

            for future in as_completed(futures):
                result = future.result()
                self.__progress_total = self.__progress_total + 1
                for content in result.search("Contents"):
                    if content:
                        all_files.append(content.get('Key'))

            all_files = [file for file in all_files if file]  # Remove None if exists.

            self.__progress_total = 0

            log.info(f"List completed in {round(time.time() - start_time, 2)} seconds.")
            return sorted(list(set(all_files)))  # Sort, dedupe, and return

    def file_matches(self, bucket_name: str, s3_path: str, search_string: str, ignore_case=False) \
            -> Union[S3MatchResult]:
        """
        Raises an ValueError if the file provided is of an unsearchable type.

        Searches a file in S3 for the matching search_string. The search_string may provide wildcards (*)

        :param bucket_name: bucket of s3 file
        :param s3_path: path to file in s3 bucket
        :param search_string: String to match on. For instance: "foo" or "foo*bar"
        :return: Hydrated S3MatchResult object
        """
        file = self._s3_resource.Object(bucket_name, s3_path)
        file_suffix_searchable = True in [file.key.lower().endswith(item) for item in self._SEARCHABLE_SUFFIXES]
        file_type_searchable = file.content_type.lower() in self._SEARCHABLE_CONTENT_TYPES

        if not file_type_searchable and not file_suffix_searchable:
            return S3MatchResult(s3_path, None, None, skipped=True)

        previous_line = ''
        for line in file.get()['Body'].iter_lines(chunk_size=4096):
            flags = re.IGNORECASE if ignore_case else 0
            current_line = line.decode()
            match_line = previous_line + current_line
            result = re.match(f'.*({search_string.replace("*", ".*")}).*', match_line, flags=flags)

            if result:
                matching_string = result[1]
                context_result = re.match(f'.*(.{{20}})({search_string.replace("*", ".*")})(.{{20}}).*',
                                          match_line, flags=flags)
                if context_result:
                    matching_context = f'{context_result[1]} {context_result[2]} {context_result[3]}'
                    return S3MatchResult(s3_path, matching_string, matching_context)
                else:
                    return S3MatchResult(s3_path, matching_string, None)
            else:
                previous_line = current_line

        return S3MatchResult(s3_path, None, None)

    def files_match(self, bucket_name: str, s3_paths: List[str], search_string: str, ignore_case=False) \
            -> List[S3MatchResult]:
        """
        For batch submitting match requests.
        """
        results = []
        for idx, path in enumerate(s3_paths):
            results.append(self.file_matches(bucket_name, path, search_string, ignore_case=ignore_case))
            with self.lock:
                self.__progress_total = self.__progress_total + 1

        return results

    def get_file_stream(self, bucket: str, s3_path: str) -> StreamingBody:
        """
        Returns a file stream object representing a file stream from S3.
        :param bucket: S3 bucket
        :param s3_path: Object path in S3
        :return: StreamingBody resouruce.
        """
        return self._s3_resource.Object(bucket, s3_path).get()['Body']

    def download_file(self, bucket: str, s3_path: str, destination: str):
        """
        Download a file locally
        :param destination: Local file path to save file to.
        """

        self._s3_resource.Object(bucket, s3_path).download_file(destination)



from typing import List


class CollectionUtils:
    @staticmethod
    def printable_set(frozen_set: frozenset):
        return f"{frozen_set}".replace("frozenset(", "").replace(")", "").replace("{", "").replace("}", "").replace("'",
                                                                                                                    "")

    @staticmethod
    def chunk_list(lst: List, chunk_size: int) -> List[List]:
        """Yield successive n-sized chunks from l."""
        for i in range(0, len(lst), chunk_size):
            yield lst[i:i + chunk_size]


import re

from figcli.config import *
from botocore.exceptions import ClientError
from prompt_toolkit import prompt
from prompt_toolkit.completion import WordCompleter

from figcli.commands.config.get import Get
from figcli.commands.config_context import ConfigContext
from figcli.commands.types.config import ConfigCommand
from figcli.data.dao.ssm import SsmDao
from figcli.input import Input
from figcli.svcs.observability.anonymous_usage_tracker import AnonymousUsageTracker
from figcli.svcs.observability.version_tracker import VersionTracker
from figcli.utils.utils import Utils
from figcli.views.rbac_limited_config import RBACLimitedConfigView


class Put(ConfigCommand):

    def __init__(self, ssm_init: SsmDao, colors_enabled: bool, config_context: ConfigContext,
                 config_view: RBACLimitedConfigView, get: Get):
        super().__init__(put, colors_enabled, config_context)
        self._ssm = ssm_init
        self._utils = Utils(colors_enabled)
        self._config_view = config_view
        self._get = get
        self._source_key = Utils.attr_if_exists(copy_from, config_context.args)

        self._select_name = [
            ('class:', 'Please input a PS Name: ')
        ]

        self._FILE_PREFIX = "file://"

    def _load_file(self, file_path: str) -> str:
        try:
            with open(file_path, 'r') as file:
                return file.read()
        except FileNotFoundError:
            print(
                f"Provided file path: {file_path} is invalid. No file found.")
            exit(1)

    def put_param(self, key=None, loop=False, display_hints=True) -> None:
        """
        Allows a user to define a PS name and add a new parameter at that named location. User will be prompted for a
        value, desc, and whether or not the parameter is a secret. If (Y) is selected for the secret, will encrypt the
        value with the appropriately mapped KMS key with the user's role.

        :param key: If specified, the user will be prompted for the specified key. Otherwise the user will be prompted
                    to specify the PS key to set.
        :param loop: Whether or not to continually loop and continue prompting the user for more keys.
        :param display_hints: Whether or not to display "Hints" to the user. You may want to turn this off if you are
                              looping and constantly calling put_param with a specified key.
        """

        value, desc, notify, put_another = True, None, False, True

        if display_hints:
            print(f"{self.c.fg_bl}Hint:{self.c.rs} To upload a file's contents, pass in `file:///path/to/your/file` "
                  f"in the value prompt.")

        while put_another:
            try:

                if not key:
                    key = Input.input('Please input a PS Name: ', completer=self._config_view.get_config_completer())

                if self._source_key:
                    plain_key = '/'.join(key.strip('/').split('/')[2:])
                    source_key = f'{self._source_key}/{plain_key}'
                    orig_value, orig_description = self._get.get_val_and_desc(source_key)
                else:
                    orig_description = ''
                    orig_value = ''

                value = Input.input(f"Please input a value for {key}: ", default=orig_value if orig_value else '')

                if value.lower().startswith(self._FILE_PREFIX):
                    value = self._load_file(value.replace(self._FILE_PREFIX, ""))

                existing_desc = self._ssm.get_description(key)
                desc = Input.input(f"Please input an optional description: ", optional=True,
                              default=existing_desc if existing_desc else orig_description if orig_description else '')

                is_secret = Input.is_secret()
                parameter_type, kms_id = SSM_SECURE_STRING if is_secret else SSM_STRING, None
                if is_secret:
                    valid_keys = self._config_view.get_authorized_kms_keys()
                    if len(valid_keys) > 1:
                        key_name = Input.select_kms_key(valid_keys)
                    else:
                        key_name = valid_keys[0]

                    kms_id = self._config_view.get_authorized_key_id(key_name, self.run_env)

                notify = True

                self._ssm.set_parameter(key, value, desc, parameter_type, key_id=kms_id)
                if key not in self._config_view.get_config_completer().words:
                    self._config_view.get_config_completer().words.append(key)

            except ClientError as e:
                if "AccessDeniedException" == e.response['Error']['Code']:
                    print(f"\n\nYou do not have permissions to put a new config value at the path:"
                          f" {self.c.fg_bl}{key}{self.c.rs}")
                    print(f"Your role has access to the following namespaces: "
                          f"{self.c.fg_bl}{self._config_view.get_authorized_namespaces()}{self.c.rs}")
                    print(f"{self.c.fg_rd}Error message: {e.response['Error']['Message']}{self.c.rs}")
                else:
                    print(f"{self.c.fg_rd}Exception caught attempting to add config: {e}{self.c.rs}")

            print()
            if loop:
                to_continue = input(f"\nAdd another? (y/N): ")
                put_another = True if to_continue.lower() == 'y' else False
                key = None
            else:
                put_another = False

    @VersionTracker.notify_user
    @AnonymousUsageTracker.track_command_usage
    def execute(self):
        self.put_param(loop=True)



import pexpect
from figcli.test.cli.config import *
from figcli.test.cli.figgy import FiggyTest
from figcli.test.cli.dev.put import DevPut
from figcli.test.cli.dev.get import DevGet
from figcli.test.cli.dev.delete import DevDelete
from figcli.config import *
from figcli.utils.utils import *
import time

KEY_UP = 'k'
KEY_DOWN = 'j'
KEY_PATH = '/shared/aaa/aa'


class DevBrowse(FiggyTest):

    def __init__(self, extra_args="", key_down_to_shared=1):
        super().__init__(None, extra_args=extra_args)
        self.key_down_to_shared = key_down_to_shared

    def run(self):
        self.step(f"Testing browse for {param_1}")
        self.browse()

    def _cleanup(self):
        delete = DevDelete(extra_args=self.extra_args)
        delete.delete(KEY_PATH, delete_another=False)

    def _setup(self):
        put = DevPut(extra_args=self.extra_args)
        put.add(KEY_PATH, DELETE_ME_VALUE, param_1_desc, add_more=False)

    def _validate_delete(self, key, value):
        print(f"Validating successfully deletion of {key}")
        get = DevGet(extra_args=self.extra_args)
        get.get(key, value, get_more=False, expect_missing=True)
        print("Delete success validated.")

    ## Get through browse, then delete
    def browse(self):
        self._setup()
        self.step("Sleeping for 60 to ensure the cache gets populated with the new /shared value")
        time.sleep(60)
        print(f"Getting {KEY_PATH} through browse...")
        # Get Value
        child = pexpect.spawn(f'{CLI_NAME} config {Utils.get_first(browse)} --env {DEFAULT_ENV} '
                              f'--skip-upgrade {self.extra_args}',
                              timeout=10, encoding='utf-8')

        for i in range(0, self.key_down_to_shared):
            child.send(KEY_DOWN)
            time.sleep(.02)

        child.send('e')
        child.send(KEY_DOWN)
        child.send(KEY_DOWN)
        child.send('e')
        child.send(KEY_DOWN)
        child.send('s')
        child.sendcontrol('n')  # <-- sends TAB
        child.sendcontrol('m')  # <-- Sends ENTER
        child.expect(f'.*{DELETE_ME_VALUE}.*')

        self.step("Get success. Deleting through browse.")
        # Delete Value
        child = pexpect.spawn(f'{CLI_NAME} config {Utils.get_first(browse)} --env {DEFAULT_ENV} '
                              f'--skip-upgrade {self.extra_args}',
                              timeout=10, encoding='utf-8')

        for i in range(0, self.key_down_to_shared):
            child.send(KEY_DOWN)
            time.sleep(.02)

        child.send('e')
        child.send(KEY_DOWN)
        child.send(KEY_DOWN)
        child.send('e')
        child.send(KEY_DOWN)
        child.send('d')
        child.sendcontrol('n')  # <-- sends TAB
        child.sendcontrol('m')  # <-- Sends ENTER
        child.expect(f'.*{KEY_PATH}.*')
        child.sendline('y')
        child.expect(f'.*{KEY_PATH}.*deleted successfully.*')

        self.step("Delete success!")
        self._validate_delete(KEY_PATH, DELETE_ME_VALUE)



import requests
from abc import ABC
import re
from typing import List
from figcli.commands.help_context import HelpContext
from figcli.commands.types.help import HelpCommand
from figcli.config import *
from figcli.input.input import Input, Utils
from figcli.models.assumable_role import AssumableRole
from figcli.models.defaults.defaults import CLIDefaults
from figcli.models.defaults.provider import Provider
from figcli.models.role import Role
from figcli.svcs.cache_manager import CacheManager
from figcli.svcs.config_manager import ConfigManager
from figcli.svcs.observability.anonymous_usage_tracker import AnonymousUsageTracker
from figcli.svcs.observability.version_tracker import VersionTracker
from figcli.svcs.setup import FiggySetup
from figcli.svcs.auth.provider.provider_factory import SessionProviderFactory
from figcli.models.sandbox.login_response import SandboxLoginResponse

from figcli.utils.awscli import AWSCLIUtils


class Login(HelpCommand, ABC):
    """
    Log the user into every possible environment they have access to. Sessions are cached.
    This improves figgy performance throughout the day.
    """

    def __init__(self, help_context: HelpContext, figgy_setup: FiggySetup):
        super().__init__(login, Utils.not_windows(), help_context)
        self._setup = figgy_setup
        self._defaults: CLIDefaults = figgy_setup.get_defaults()
        self._utils = Utils(self._defaults.colors_enabled)
        self.example = f"\n\n{self.c.fg_bl}{CLI_NAME} {Utils.get_first(login)} \n" \
                       f"{self.c.rs}{self.c.fg_yl}  --or--{self.c.rs}\n" \
                       f"{self.c.fg_bl}{CLI_NAME} {Utils.get_first(login)} {Utils.get_first(sandbox)}{self.c.rs}"

    def login(self):
        self._utils.validate(self._defaults.provider.name in Provider.names(),
                             f"You cannot login until you've configured Figgy. Please run `{CLI_NAME}` --configure")
        provider = SessionProviderFactory(self._defaults).instance()
        assumable_roles: List[AssumableRole] = provider.get_assumable_roles()
        print(f"{self.c.fg_bl}Found {len(assumable_roles)} possible logins. Logging in...{self.c.rs}")

        for role in assumable_roles:
            print(f"Login successful for {role.role} in environment: {role.run_env}")
            provider.get_session_and_role(role, False)

        print(f"{self.c.fg_gr}Login successful. All sessions are cached.{self.c.rs}")

    def login_sandbox(self):
        Utils.wipe_vaults() or Utils.wipe_defaults() or Utils.wipe_config_cache()
        print(f"{self.c.fg_bl}Logging you into the Figgy Sandbox environment.{self.c.rs}")
        user = Input.input("Please input a user name: ")
        colors = Input.select_enable_colors()
        print()

        role = Input.select("Please select a role to impersonate: ", valid_options=SANDBOX_ROLES)
        params = {'role': role, 'user': user}
        result = requests.get(GET_SANDBOX_CREDS_URL, params=params)

        if result.status_code != 200:
            self._utils.error_exit("Unable to get temporary credentials from the Figgy sandbox. If this problem "
                                   f"persists please notify us on our GITHUB: {FIGGY_GITHUB}")

        data = result.json()
        response = SandboxLoginResponse(**data)
        AWSCLIUtils.write_credentials(access_key=response.AWS_ACCESS_KEY_ID, secret_key=response.AWS_SECRET_ACCESS_KEY,
                                      token=response.AWS_SESSION_TOKEN, region=FIGGY_SANDBOX_REGION,
                                      profile_name=FIGGY_SANDBOX_PROFILE, color=self.c)

        defaults = CLIDefaults.sandbox(user=user, role=role, colors=colors)
        self._setup.save_defaults(defaults)

        config_mgr = ConfigManager.figgy()
        config_mgr.set(Config.Section.Bastion.PROFILE, FIGGY_SANDBOX_PROFILE)
        defaults = self._setup.configure_extras(defaults)
        defaults = self._setup.configure_roles(current_defaults=defaults, role=Role(role))
        defaults = self._setup.configure_figgy_defaults(defaults)
        self._setup.save_defaults(defaults)

        print(f"\n{self.c.fg_gr}Login successful. Your sandbox session will last for{self.c.rs} "
              f"{self.c.fg_bl}1 hour.{self.c.rs}")

        print(
            f"\nIf your session expires, feel free to rerun `{CLI_NAME} login sandbox` to get another sandbox session. "
            f"\nAll previous figgy sessions have been disabled, you'll need to run {CLI_NAME} "
            f"--configure to leave the sandbox.")

    @VersionTracker.notify_user
    @AnonymousUsageTracker.track_command_usage
    def execute(self):
        if self.context.command == login:
            self.login()
        elif self.context.command == sandbox:
            Utils.wipe_vaults() or Utils.wipe_defaults()
            self.login_sandbox()



from figcli.models.restore_config import RestoreConfig
from typing import List
import json
import datetime
from decimal import *
from figcli.config import *

class ParameterHistory:
    """
    Represents the history of a single configuration with a single name.
    """

    def __init__(self):
        self.history: List[RestoreConfig] = []
        self.name = None

    @staticmethod
    def instance(cfg: RestoreConfig):
        history = ParameterHistory()
        history.add(cfg)
        return history

    def cfgs_before(self, ps_time: datetime.datetime):
        time = Decimal(ps_time.timestamp() * 1000)
        cfgs: List[RestoreConfig] = []
        for cfg in self.history:
            if cfg.ps_time < time and cfg.ps_action == SSM_PUT:
                cfgs.append(cfg)

        return sorted(cfgs, key=lambda x: x.ps_time)

    def cfg_at(self, ps_time: datetime.datetime):
        time = Decimal(ps_time.timestamp() * 1000)
        previous_cfg: RestoreConfig = None
        restore_cfg = None
        for cfg in self.history:
            if previous_cfg and previous_cfg.ps_time < time < cfg.ps_time:
                restore_cfg = previous_cfg
                break
            else:
                previous_cfg = cfg

        if not restore_cfg:
            restore_cfg = previous_cfg

        return restore_cfg

    def __eq__(self, other):
        return hash(self.__dict__) == hash(other.__dict__)

    def __hash__(self):
        return hash(self.__dict__)

    def __str__(self):
        return f"{self.__dict__}"

    def add(self, config: RestoreConfig):
        if not self.name:
            self.name = config.ps_name

        # Add and resort
        self.history.append(config)
        self.history = sorted(self.history, key=lambda x: x.ps_time)



from abc import ABC, abstractmethod
from typing import List

from pexpect.exceptions import TIMEOUT
from figcli.utils.utils import Utils
import sys


class FiggyTest(ABC):

    def __init__(self, child, extra_args=""):
        self.c = Utils.default_colors()
        self.extra_args = extra_args

        if child:
            c = Utils.default_colors()
            print(f"{c.fg_yl}Testing command: {child.args}{c.rs}")
            self._child = child
            self._child.logfile = sys.stdout

    @abstractmethod
    def run(self):
        pass

    def expect_multiple(self, regexes: List[str]):
        print(f'Expecting: {regexes}')
        return self._child.expect(regexes)

    def expect(self, regex: str):
        print(f'Expecting: {regex}')
        self._child.expect(regex)

    def sendline(self, line: str):
        print(f'Sending: {line}')
        self._child.sendline(line)

    def step(self, step_msg: str):
        print(f"{self.c.fg_bl}-----------------------------------------{self.c.rs}")
        print(f"{self.c.fg_yl} STEP: {step_msg}{self.c.rs}")
        print(f"{self.c.fg_bl}-----------------------------------------{self.c.rs}")


import pexpect
import sys
from figcli.test.cli.config import *
from figcli.test.cli.figgy import FiggyTest
from figcli.config import *
from figcli.utils.utils import *
from figcli.test.cli.dev.put import DevPut
from figcli.test.cli.dev.delete import DevDelete
import time


class DevDump(FiggyTest):

    def __init__(self, extra_args=""):
        super().__init__(None, extra_args=extra_args)

    def run(self):
        put = DevPut(extra_args=self.extra_args)

        # Use a number > 10 so paging is tested
        minimum, maximum = 1, 12
        put.add(param_1, param_1_val, param_1_desc, add_more=True)

        for i in range(minimum, maximum):
            more = i < maximum - 1
            put.add_another(f'{param_1}-{i}', param_1_val, f'{param_1_desc}-{i}', add_more=more)

        child = pexpect.spawn(f'{CLI_NAME} config {Utils.get_first(dump)} --env {DEFAULT_ENV} {self.extra_args}'
                              f' --skip-upgrade',
                              encoding='utf-8', timeout=7)

        self.step(f"Testing `{CLI_NAME} config {Utils.get_first(dump)} --env {DEFAULT_ENV}`")
        child.expect('.*to dump from.*')
        child.sendline(dump_prefix)
        child.expect(f'.*{param_1}-{minimum}.*{param_1}-{maximum-1}.*')
        print(f"Dump was successful.")

        delete = DevDelete(extra_args=self.extra_args)
        delete.delete(param_1, delete_another=True)
        for i in range(minimum, maximum):
            delete.delete(f'{param_1}-{i}', delete_another=i < maximum - 1)



import logging
from typing import Set, List

from figcli.data.dao.config import ConfigDao
from figcli.data.models.config_item import ConfigState, ConfigItem
from figcli.models.run_env import RunEnv
from figcli.svcs.cache_manager import CacheManager
from figcli.utils.utils import Utils

log = logging.getLogger(__name__)


# Todo, lots more from SSMDao should be moved here.
class ConfigService:
    __CACHE_REFRESH_INTERVAL = 60 * 60 * 24 * 7 * 1000  # 1 week in MS
    """
    Service level class for interactive with config resources.

    Currently contains some business logic to leverage a both local (filesystem) & remote (dynamodb)
    cache for the fastest possible lookup times.
    """
    _PS_NAME_CACHE_KEY = 'parameter_names'

    def __init__(self, config_dao: ConfigDao, cache_mgr: CacheManager, run_env: RunEnv):
        self._config_dao = config_dao
        self._cache_mgr = cache_mgr
        self._run_env = run_env

    def get_root_namespaces(self) -> List[str]:
        all_params = self.get_parameter_names()
        return sorted(list(set([f"/{p.split('/')[1]}" for p in all_params])))

    @Utils.trace
    def get_parameter_names(self) -> Set[str]:
        """
        Looks up local cached configs, then queries new config names from the remote cache, merges the two, and
        finally updates the local cache. This ensures very fast bootstrap times b/c querying thousands of parameter
        names from a remote cache can a bit too much time. `figgy` does not accept slow performance.

        :return: Set[str] names of all configs stored in ParameterStore.
        """
        cache_key = f'{self._run_env.env}-{self._PS_NAME_CACHE_KEY}'

        # Find last cache full refresh date
        last_refresh = self._cache_mgr.last_refresh(cache_key)

        # Do a full refresh if cache is too old.
        if Utils.millis_since_epoch() - last_refresh > self.__CACHE_REFRESH_INTERVAL:
            configs: Set[ConfigItem] = self._config_dao.get_config_names_after(0)
            all_parameters: Set[str] = set([x.name for x in configs if x.state == ConfigState.ACTIVE])
            self._cache_mgr.write(cache_key, all_parameters)
        else:

            # Get items from cache
            last_write, cached_contents = self._cache_mgr.get(cache_key)

            # Find new items added to remote cache table since last local cache write
            updated_items: Set[ConfigItem] = self._config_dao.get_config_names_after(last_write)

            # Add new names to cache
            added_names, deleted_names = set(), set()
            for item in updated_items:
                if item.state is ConfigState.ACTIVE:
                    log.info(f"ADDING NEW: {item}")
                    added_names.add(item.name)
                elif item.state is ConfigState.DELETED:
                    log.info(f"ADDING DELETED: {item}")
                    deleted_names.add(item.name)
                else:
                    # Default to add if no state set
                    added_names.add(item.name)

            self._cache_mgr.append(cache_key, added_names)
            self._cache_mgr.delete(cache_key, deleted_names)

            log.debug(f"Cached: {cached_contents}")
            log.debug(f"Cached: {deleted_names}")
            log.debug(f"Cached: {added_names}")

            all_parameters = set(cached_contents) - deleted_names | added_names

        return all_parameters



import pexpect
from figcli.test.cli.config import *
from figcli.test.cli.figgy import FiggyTest
from figcli.config import *
from figcli.utils.utils import *
import sys


class DevPut(FiggyTest):
    def __init__(self, extra_args=""):
        print(f"Testing `figgy config {Utils.get_first(put)} --env {DEFAULT_ENV}`")
        super().__init__(pexpect.spawn(f'{CLI_NAME} config {Utils.get_first(put)} '
                                    f'--env {DEFAULT_ENV} --skip-upgrade {extra_args}', timeout=7, encoding='utf-8'))

    def run(self):
        self.step(f"Testing PUT for {param_1}")
        self.add(param_1, param_1_val, param_1_desc)

    def add(self, key, value, desc, add_more=False):
        self.expect('.*Please input a PS Name.*')
        self.sendline(key)
        self.expect('.*Please input a value.*')
        self.sendline(value)
        self.expect('.*Please input an optional.*')
        self.sendline(desc)
        self.expect('.*secret?.*')
        self.sendline('n')
        self.expect('.*another.*')
        if add_more:
            self.sendline('y')
        else:
            self.sendline('n')

    def add_another(self, key, value, desc, add_more=True):
        print(f"Adding another: {key} -> {value}")
        self.expect('.*PS Name.*')
        self.sendline(key)
        self.expect('.*Please input a value.*')
        self.sendline(value)
        self.expect('.*Please input an optional.*')
        self.sendline(desc)
        self.expect('.*secret?.*')
        self.sendline('n')
        self.expect('.*Add another.*')
        if add_more:
            self.sendline('y')
        else:
            self.sendline('n')



import re
from setuptools import setup, find_packages
import platform
import os
# Platform Constants
LINUX, MAC, WINDOWS = "Linux", "Darwin", "Windows"

SHORT_DESCRIPTION = "Figgy is a free and opensource serverless application config framework designed to bring " \
                    "simplicity, security, and resilience to application config management. Figgy is built on top of" \
                    " AWS ParameterStore and leverages native AWS constructs such as AWS IAM, KMS, among other " \
                    "services to ensure a simple and elegant integration with your AWS environment."

with open('figcli/config/constants.py') as file:
    contents = file.read()
    VERSION = re.search(r'^VERSION\s*=\s*["\'](.*)["\']', contents, re.MULTILINE)
    GITHUB = re.search(r'^FIGGY_GITHUB\s*=\s*["\'](.*)["\']', contents, re.MULTILINE)

VERSION = VERSION.group(1)
GITHUB = GITHUB.group(1)
FIGGY_WEBSITE = "https://figgy.dev"

base_requirements = [
        "boto3 >= 1.13.19",
        "prompt_toolkit == 2.0.7",
        "sty >= 1.0.0b8",
        "click >= 7.1.2",
        "tqdm >= 4.46.0",
        "npyscreen >= 4.10.5",
        "pexpect >= 4.8.0",
        "mfa >= 0.1.1",
        "beautifulsoup4 >= 4.9.1",
        "keyring >= 21.2.1",
        "keyrings.alt >= 3.4.0",
        "tabulate >= 0.8.7",
        "jsonpickle >= 1.4.1",
        "urllib3 >= 1.25.7",
        "pyotp >= 2.3.0",
        "pydantic >= 1.5.1",
        "python-u2flib-host",
        "pycryptodome>=3.9.7",
]

windows_requirements = [
    "pyreadline>=2.1",
    "windows-curses>=2.1.0",
    "pywin32",
]

linux_requirements = [
]

darwin_requriements = [
]

if platform.system() == WINDOWS:
    requirements = base_requirements + windows_requirements
elif platform.system() == LINUX:
    requirements = base_requirements + linux_requirements
elif platform.system() == MAC:
    requirements = base_requirements + darwin_requriements
else:
    requirements = base_requirements + linux_requirements

if os.environ.get('FIGGY_TEST') == 'true':
    excludes = []
else:
    excludes = ["test"]

LONG_DESCRIPTION = """
# figgy
Cloud native config management.

Figgy is currently under active development and will be released as soon as it's ready! :)


# Figgy is not ready yet, it's still under active development :)


# **What's Figgy?**


Figgy is a **_free_** and **_opensource_** serverless application config framework designed to bring simplicity, security, and resilience to 
application config management. Figgy is built on top of AWS ParameterStore and leverages native AWS constructs such as AWS IAM, 
KMS, among other services to ensure a simple and elegant integration with your AWS environment.
<br/>

> **Never roll another application to production having forgotten to set that last pesky
config in production.**

Figgy makes it possible to **bind your code directly to configurations**. Easily break builds if configs 
are missing and application deployments are destined to fail.


> **Control user access like a champ**

Figgy makes it easy to set up and control access to across all of your AWS environments and configuration namespaces. Consider
your role types and use cases, map them up in a simple config file, and let Figgy do the rest. Audit all user activity and 
changes over time, and roll back any config or group of configurations to any point-in-time -- to the second!

> **Integrate with your SSO provider, abandon long-lived AWS Keys for good**

Figgy supports SAML based SSO integrations with multi-factor authentication. Simplify AWS access control with Figgy!

> **Feature rich CLI to speed-up your development workflow.**
![Figgy Get Browse]({{ "/assets/img/animations/home/get-browse.gif"| relative_url }})

<a name="concepts"></a>
<br/>
## **Figgy Concepts**:

- **Fig Orchard** - All con**fig**urations under all figgy-managed namespaces. 
    - **These are examples. You select your own namespaces.**
  - `/app/*`
  - `/shared/*`
  - `/dba/*`
  - `/devops/*`
  - `/sre/*`


- **Fig Tree** - A hierarchy of  under a high-level figgy-managed namespace.
    - e.g. - `/app/*`    
    
    
- **Twig** - A single hierarchy of configs under a **Fig Tree**
    - e.g. - `/app/hello-world/*`
    - Write your service IAM policies to ONLY access the *twig* namespace
    
    
- **Fig** - A single con**fig**uration stored `/under/a/named/path` and on a **Twig**
    - e.g. - `/app/hello-world/log/level`
  
> **These namespaces are all examples. You can name or Figs / Twigs / Trees, and Orchards however you like!**

<br/>
## **Why Figgy?**

*Out of the box, Figgy comes with all of these features:*

- **SSO Integrations with Google Admin Console, OKTA, and AWS (more to come)**
    - MFA is supported and encouraged
    - Figgy ONLY uses temporary credentials. Abandon all AWS access keys!

- **An elegant CLI on top of AWS ParameterStore that addresses many ParameterStore limitations:**
    - Add / Update / Delete / Edit configurations and more
    - Promote configs from lower to higher environments
    - Share secrets directly to the code that needs them. No more handing DB credentials to some middle man so they can go put them "somewhere".
    - Browse a log that tracks all config changes over time, even for deleted configs.
    - Roll back any configuration, or hierarchy of configurations to *any point in time* (to the second) in the past!
    - Combat config sprawl. Figgy will tell you if you have a config in ParameterStore that you aren't using anymore!

- **Security**
    - Create figgy 'roles' that allow different user types access to different namespaces in your configuration tree.
    - Easily control access between different configuration trees.
    - Securely share secrets between config trees
    - Track all configuration changes over time and restore changes to any point-in-time in the past!

- **Binding application configs to your code!**
    - Easily integrate your CICD process with figgy
    - **BREAK THE BUILD** if the application you're deploying is missing a required config in the environment you're
        deploying to. 
    - Give Developers confidence their code bootstrap properly if Figgy gives the thumbs-up! 
    - Easily determine application dependencies in _one place_ by looking at your application's **Fig Tree** 

- **The Figgy Vault**
    - Figgy _only_ generates temporary sessions to AWS, encrypts them, and stores them locally in your personal "Figgy Vault"
    - These temporary credentials can be used for local development by decrypting & pulling them from the vault.

- **Slack integration**
    - Get automated notifications to slack when secrets are changed or updated, and know who made them.


and a lot more!
"""

setup(
    name="figgy-cli",
    packages=find_packages(".", exclude=excludes),
    entry_points={
        "console_scripts": ['figgy = figcli.entrypoint.cli:main']
    },
    version=VERSION,
    description=SHORT_DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author="Jordan Mance",
    author_email="jordan@figgy.dev",
    url=FIGGY_WEBSITE,
    python_requires='>=3.7',
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Environment :: Console",
        'Intended Audience :: Developers',
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation",
        "Topic :: Terminals",
        "Topic :: Utilities",

    ]
)



import logging
import botocore
from figcli.utils.utils import *
from botocore.exceptions import ClientError
from typing import List, Optional
from figcli.utils.utils import Utils
from concurrent.futures import ThreadPoolExecutor
from typing import Tuple


class SsmDao:
    """
    Should be used for all queries to SSM, has built in pagination, etc.
    """
    max_results = 50  # This is the max according to api docs.

    def __init__(self, boto_ssm_client):
        self._ssm = boto_ssm_client
        self._utils = Utils(False)

    @Utils.retry
    def get_parameter_values(self, parameters: List[str], decrypt: bool = False) -> List[Dict]:
        """
        Queries values for a series of parameters, and decrypts SecureStrings if requested.
        Args:
            parameters: List[str]: Parameter names to query values for.
            decrypt: bool: True/False - Attempt to decrypt values during query.

        Returns: List[Dict] - List of Dictionaries container parameter data
        """
        results = []  # type: List[Dict]
        chunks = Utils.chunk_list(parameters, 10)
        for chunk in chunks:
            result = self._ssm.get_parameters(Names=chunk, WithDecryption=decrypt)
            if 'Parameters' in result:
                results = results + result['Parameters']

        return results

    @Utils.retry
    def get_all_parameters(self, prefixes: List[str], option: str = 'Recursive', page: str = None) -> List[dict]:
        """
        Returns all parameters under prefix. Automatically pages recursively then returns full result set
        Args:
            prefixes: List of prefixes to query. E.G. [ '/shared', '/data', '/app' ]
            existing_params: Used in recursive calls to build a total result set
            page: Used in recursive calls if more pages exist.
            option: Must be 'Recursive' or 'OneLevel' - Indicates # of levels below the prefix to recurse.
        Returns: List[dict] -> Parameter details as returned from AWS API

        """
        filters = {
                      'Key': 'Path',
                      'Option': f'{option}',
                      'Values': prefixes
                  },
        total_params = []
        if page:
            params = self._ssm.describe_parameters(ParameterFilters=filters, NextToken=page,
                                                   MaxResults=self.max_results)
        else:
            params = self._ssm.describe_parameters(ParameterFilters=filters, MaxResults=self.max_results)

            self._utils.validate(params and 'Parameters' in params,
                                 f"Failed to lookup parameters with prefix: {prefixes}")
        total_params = total_params + params['Parameters']

        if params and 'NextToken' in params:
            total_params = total_params + self.get_all_parameters(prefixes, option=option, page=params['NextToken'])

        return total_params

    @Utils.retry
    def get_parameter_details(self, name: str) -> Dict:
        """
        Get the parameter details for a single parameter. This is extremely inefficient, due to how the AWS API is
        designed, you should probably avoid this.
        :param name: The name of the parameter
        :return: Dict -> Parameter details as returned from AWS API or an empty Dict {} if nothing is found.
        """
        filters = {
            'Key': 'Name',
            'Values': [name]
        }
        next_token = True
        while next_token:
            if isinstance(next_token, str):
                result = self._ssm.describe_parameters(Filters=[filters], NextToken=next_token,
                                                       MaxResults=self.max_results)
            else:
                result = self._ssm.describe_parameters(Filters=[filters], MaxResults=self.max_results)

            params = result.get('Parameters')
            if params:
                return params.pop()
            else:
                next_token = result.get('NextToken')

        return {}

    @Utils.retry
    def get_parameter_with_description(self, name: str) -> Tuple[Optional[str], Optional[str]]:
        """
        Returns a parameter's value and description from its provided name. Returns `None, None` tuple
        if no parameter exists.

        This method paginates to the last page via the SSM API then grabs the last item, which will
        be the current version.

        :param name: The name of the parameter - e.g. /app/foo/bar
        :return: Tuple[value, description] - Value & Description found. None returned if no parameter exists.
        """
        try:
            next_token, result = True, {}

            while next_token:
                if isinstance(next_token, str):
                    result = self._ssm.get_parameter_history(
                        Name=name,
                        WithDecryption=True,
                        NextToken=next_token
                    )
                else:
                    result = self._ssm.get_parameter_history(
                        Name=name,
                        WithDecryption=True
                    )

                next_token = result.get('NextToken')

            history = result.get('Parameters', [])
            if history:
                current_val = history[-1]
                return current_val.get('Value'), current_val.get('Description')
            else:
                return None, None

        except ClientError as e:
            if "ParameterNotFound" == e.response['Error']['Code']:
                return None, None
            else:
                raise

    @Utils.retry
    def get_description(self, name: str) -> Union[str, None]:
        """
        Returns the description for a parameter in parameter store, or None if no description exists.
        :param name: /path/to/parameter
        :return: str - description of the specified parameter, or None if no description exists
        """

        value, desc = self.get_parameter_with_description(name)
        return desc

    @Utils.retry
    def get_all_param_names_fast(self, prefixes: List[str], submitted_prefixes: set = None, page: str = None) -> set:
        """
        Recursively gets /one/level/of/parameters at a time, over many threads to find all params faster
        Args:
            prefixes: List of prefixes to query. E.G. [ '/shared', '/data', '/app' ]
            existing_params: Used in recursive calls to build a total result set
            page: Used in recursive calls if more pages exist.
        Returns: List[dict] -> Parameter details as returned from AWS API
        """
        if submitted_prefixes is None:
            submitted_prefixes = set()

        filters_one = {'Key': 'Path', 'Option': 'Recursive', 'Values': prefixes},
        submitted_prefixes = submitted_prefixes | set(prefixes)

        if page:
            params = self._ssm.describe_parameters(ParameterFilters=filters_one, MaxResults=self.max_results,
                                                   NextToken=page)
            # print(f"Looking up with prefixes: {prefixes} - PAGED")
        else:
            params = self._ssm.describe_parameters(ParameterFilters=filters_one, MaxResults=self.max_results)
            print(f"{prefixes}")

        all_names = set()

        self._utils.validate(params and 'Parameters' in params, f"Failed to lookup parameters with prefix: {prefixes}")
        names = set(map(lambda x: x['Name'], params['Parameters']))
        all_names = all_names | names
        futures = []

        if len(names) > 0:
            for prefix in prefixes:
                next_level = set(
                    list(map(lambda x: x.replace(prefix, '').split('/')[1],
                             list(filter(lambda n:
                                         n.startswith(prefix) and
                                         len(n.replace(prefix, '').split('/')) > 2, names))))
                )
                # print(f"Next level: {next_level} from prefix: {prefix}")
                with ThreadPoolExecutor(max_workers=10) as pool:
                    for param in next_level:
                        if param not in submitted_prefixes:
                            submitted_prefixes.add(param)
                            print(f"Getting for next level {prefix}/{param}")
                            futures.append(pool.submit(self.get_all_param_names_fast,
                                                       [f'{prefix}/{param}'], submitted_prefixes=submitted_prefixes))

        if params and 'NextToken' in params:
            all_names = all_names | self.get_all_param_names_fast(prefixes, submitted_prefixes=submitted_prefixes,
                                                                  page=params['NextToken'])

        for item in futures:
            all_names = all_names | item.result()

        return all_names

    @Utils.retry
    def delete_parameter(self, key) -> None:
        """
        Deletes a parameter from PS
        Args:
            key: The PS Name - E.G. /app/demo-time/parameter/abc123

        Returns:

        """
        response = self._ssm.delete_parameter(Name=key)
        self._utils.validate(
            response and response['ResponseMetadata'] and response['ResponseMetadata']['HTTPStatusCode']
            and response['ResponseMetadata']['HTTPStatusCode'] == 200,
            f"Error deleting key: [{key}] from PS. Please try again.")

    @Utils.retry
    def get_parameter(self, key) -> Optional[str]:
        """
        Gets a parameter, returns None if parameter doesn't exist.
        Args:
            key: The PS Name - E.G. /app/demo-time/parameter/abc123

        Returns: str -> Parameter's value

        """
        try:
            parameter = self._ssm.get_parameter(Name=key, WithDecryption=True)
            return parameter['Parameter']['Value']
        except ClientError as e:
            if "ParameterNotFound" == e.response['Error']['Code']:
                return None
            else:
                raise

    @Utils.retry
    def get_parameter_encrypted_by_version(self, key: str, ps_version: int) -> Optional[str]:
        try:
            parameter_history = self._ssm.get_parameter_history(
                Name=key,
                WithDecryption=False,
                MaxResults=10,
            )
            if len(parameter_history["Parameters"]) > 0:
                parameter = list(filter(lambda x: x["Version"] == ps_version, parameter_history["Parameters"]))
                if parameter:
                    return parameter[0]["Value"]
            else:
                return None
        except ClientError as e:
            if "ParameterNotFound" == e.response['Error']['Code']:
                return None
            else:
                raise

    @Utils.retry
    def get_parameter_encrypted(self, key):
        """
            Returns the parameter without decrypting the value. If parameter isn't encrypted, it returns the value.
        Args:
            key: The PS Name - E.G. /app/demo-time/parameter/abc123

        Returns: str -> encrypted string value of an encrypted parameter.

        """
        try:
            parameter = self._ssm.get_parameter(Name=key, WithDecryption=False)
            return parameter['Parameter']['Value']
        except ClientError as e:
            if "ParameterNotFound" == e.response['Error']['Code']:
                return None
            else:
                raise

    @Utils.retry
    def set_parameter(self, key, value, desc, type, tags=[], key_id=None) -> None:
        """
        Sets a parameter in PS.
        Args:
            key: The PS Name - E.G. /app/demo-time/parameter/abc123
            value: Value to set
            desc: Description
            type: SecureString or String
            key_id: KMS Key Id to use for encryption if SecureString
        """
        # Doesn't belong in DAO. Move to SVC layer when that refactor happens.
        # tags = tags + [{
        #     'Key': 'managed_by',
        #     'Value': PROJECT_NAME
        #     },
        # ]

        if key_id:
            self._ssm.put_parameter(
                Name=key,
                Description=desc,
                Value=value,
                Overwrite=True,
                Type=type,
                KeyId=key_id,
                # Tags=tags
            )
        else:
            self._ssm.put_parameter(
                Name=key,
                Description=desc,
                Value=value,
                Overwrite=True,
                Type=type,
                # Tags=tags
            )



from figcli.config import *
import boto3
from boto3.dynamodb.conditions import Key, Attr
from typing import List


class TFEDao:
    def __init__(self, mgmt_session: boto3.Session):
        self._mgmt_session = mgmt_session
        self._tfe_cache = self._mgmt_session.resource('dynamodb').Table(TFE_CACHE_TABLE_NAME)

    def get_all_workspaces(self, start_key: str=None) -> List[str]:
        if start_key:
            result = self._tfe_cache.scan(ExclusiveStartKey=start_key)
        else:
            filter_exp = Attr(TFE_TYPE_ATTR_NAME).eq('workspace')
            result = self._tfe_cache.scan(FilterExpression=filter_exp)

        items: List = result["Items"] if result["Items"] else []

        if len(items) == 0:
            return items

        results: List[str] = list(map(lambda x: x["resource_name"], items))

        if 'LastEvaluatedKey' in result:
            results = results + self.get_all_workspaces(start_key=result['LastEvaluatedKey'])

        return results



# THIS MUST MATCH devops.ci.lambdas/lambdas/python/parameter_store_auditor/config.py's value!!!
### For PS items stored with this value, we will auto-clean them up from our audit table. Used for automated E2E testing.
DELETE_ME_VALUE = 'DELETE_ME'   ### <-- use this for ALL VALUES
MFA_USER_ENV_KEY = 'MFA_USER'
MFA_SECRET_ENV_KEY = 'MFA_SECRET'

# Env vars
GOOGLE_SSO_USER = 'GOOGLE_SSO_USER'
GOOGLE_SSO_PASSWORD = 'GOOGLE_SSO_PASSWORD'
GOOGLE_IDP_ID = 'GOOGLE_IDP_ID'
GOOGLE_SP_ID = 'GOOGLE_SP_ID'

OKTA_SSO_USER = 'OKTA_SSO_USER'
OKTA_SSO_PASSWORD = 'OKTA_SSO_PASSWORD'
OKTA_EMBED_URL = 'OKTA_EMBED_URL'
OKTA_MFA_SECRET = 'OKTA_MFA_SECRET'


# Variables
param_1 = '/shared/test/automated_test/param_1'
param_test_prefix = '/shared/test2/automated_test/'
dump_prefix = '/shared/test/automated_test/'
data_param_1 = '/data/test/automated_test/parm_1'
devops_param_1 = '/devops/test/automated_test/param_1'

# Values
param_1_val = DELETE_ME_VALUE
data_param_1_val = DELETE_ME_VALUE
devops_param_1_val = DELETE_ME_VALUE


# Desc
param_1_desc = 'desc1'
data_param_1_desc = 'datadesc1'
devops_param_1_desc = 'devopsdesc1'


# Share destination
automated_test_dest_1 = '/app/automated-test/dest/1'


# Others
DEFAULT_ENV = 'stage'


import time
from dataclasses import dataclass

from figcli.models.defaults.defaults import CLIDefaults
from figcli.models.sso.okta.okta_session import OktaSession
from figcli.utils.utils import *
import logging
import requests
from figcli.models.sso.okta.okta_auth import OktaAuth

log = logging.getLogger(__name__)


# Pulled from / tweaked -> https://github.com/jmhale/okta-awscli/blob/master/oktaawscli/aws_auth.py

class OktaPrimaryAuth(OktaAuth):
    def __init__(self, defaults: CLIDefaults, password: str, mfa: str):
        super().__init__(defaults)
        self._user = defaults.user
        self._password = password
        self.https_base_url = f"https://{defaults.provider_config.base_url}"
        self.totp_token = mfa
        self.factor = self.factor_type
        self._session: OktaSession = None

    def get_session(self):
        return OktaSession(session_id=self._get_session().session_id,
                           session_token=self._get_session().session_token)

    def to_json(self) -> str:

        return json.dumps({
            OKTA_SESSION_TOKEN_CACHE_KEY: self.get_session().session_token,
            OKTA_SESSION_ID_CACHE_KEY: self.get_session().session_id
        })

    def _get_session(self) -> OktaSession:
        """ Performs primary auth against Okta """
        if self._session:
            return self._session

        auth_data = {
            "username": self._user,
            "password": self._password
        }
        resp = requests.post(self.https_base_url + '/api/v1/authn', json=auth_data)
        resp_json = resp.json()
        if 'status' in resp_json:
            if resp_json['status'] == 'MFA_REQUIRED':
                factors_list = resp_json['_embedded']['factors']
                state_token = resp_json['stateToken']
                session_token = self.verify_mfa(factors_list, state_token)
            elif resp_json['status'] == 'SUCCESS':
                session_token = resp_json['sessionToken']
            elif resp_json['status'] == 'MFA_ENROLL':
                log.warning("MFA not enrolled. Cannot continue. Please enroll a MFA factor in the Okta Web "
                            "UI first!")
                raise InvalidSessionError("Unable to auth through Okta.")
        elif resp.status_code != 200:
            log.error(resp_json['errorSummary'])
            raise InvalidSessionError("Unable to auth through Okta.")
        else:
            print("Error creating OKTA session. Perhaps invalid MFA. Please retry.")
            log.error(resp_json)
            raise InvalidSessionError("Unable to auth through Okta.")

        session_id = self.get_session_id(session_token)
        # Cache session for subsequent requests.
        self._session = OktaSession(session_token=session_token, session_id=session_id)

        return self._session

    def verify_mfa(self, factors_list, state_token) -> str:
        """ Performs MFA auth against Okta """

        supported_factor_types = ["token:software:totp", "push"]

        supported_factors = []
        for factor in factors_list:
            if factor['factorType'] in supported_factor_types:
                supported_factors.append(factor)
            else:
                log.info("Unsupported factorType: %s" %
                         (factor['factorType'],))

        supported_factors = sorted(supported_factors,
                                   key=lambda factor: (
                                       factor['provider'],
                                       factor['factorType']))
        if len(supported_factors) == 1:
            session_token = self.verify_single_factor(
                supported_factors[0], state_token)
        elif len(supported_factors) > 0:
            if not self.factor:
                print("Registered MFA factors:")
            for index, factor in enumerate(supported_factors):
                factor_type = factor['factorType']
                factor_provider = factor['provider']

                if factor_provider == "GOOGLE":
                    factor_name = "Google Authenticator"
                elif factor_provider == "OKTA":
                    if factor_type == "push":
                        factor_name = "Okta Verify - Push"
                    else:
                        factor_name = "Okta Verify"
                else:
                    factor_name = "Unsupported factor type: %s" % factor_provider

                if self.factor:
                    if self.factor == factor_provider:
                        factor_choice = index
                        log.info("Using pre-selected factor choice \
                                         from ~/.okta-aws")
                        break
                else:
                    print("%d: %s" % (index + 1, factor_name))
            if not self.factor:
                factor_choice = int(input('Please select the MFA factor: ')) - 1
            log.info("Performing secondary authentication using: %s" %
                     supported_factors[factor_choice]['provider'])
            session_token = self.verify_single_factor(supported_factors[factor_choice],
                                                      state_token)
        else:
            print("MFA required, but no supported factors enrolled! Exiting.")
            exit(1)
        return session_token

    def verify_single_factor(self, factor, state_token):
        """ Verifies a single MFA factor """
        req_data = {
            "stateToken": state_token
        }

        log.debug(factor)
        if factor['factorType'] == 'token:software:totp':
            if self.totp_token:
                log.debug("Using TOTP token from command line arg")
                req_data['answer'] = self.totp_token
            else:
                req_data['answer'] = input('Enter MFA token: ')

        post_url = factor['_links']['verify']['href']
        resp = requests.post(post_url, json=req_data)
        resp_json = resp.json()
        if 'status' in resp_json:
            if resp_json['status'] == "SUCCESS":
                return resp_json['sessionToken']
            elif resp_json['status'] == "MFA_CHALLENGE" and factor['factorType'] != 'u2f':
                print("Waiting for push verification...")
                while True:
                    resp = requests.post(
                        resp_json['_links']['next']['href'], json=req_data)
                    resp_json = resp.json()
                    if resp_json['status'] == 'SUCCESS':
                        return resp_json['sessionToken']
                    elif resp_json['factorResult'] == 'TIMEOUT':
                        print("Verification timed out")
                        exit(1)
                    elif resp_json['factorResult'] == 'REJECTED':
                        print("Verification was rejected")
                        exit(1)
                    else:
                        time.sleep(0.5)
        return None

    def get_session_id(self, session_token):
        """ Gets a session cookie from a session token """
        data = {"sessionToken": session_token}
        resp = requests.post(
            self.https_base_url + '/api/v1/sessions', json=data).json()

        if 'id' not in resp:
            raise InvalidSessionError("Failed to get session from OKTA.")

        return resp['id']



import sys

import pexpect
from figcli.test.cli.config import *
from figcli.test.cli.figgy import FiggyTest
from figcli.test.cli.dev.put import DevPut
from figcli.test.cli.dev.get import DevGet
from figcli.test.cli.dev.delete import DevDelete
from figcli.config import *
from figcli.utils.utils import *
import time


#Todo get this working...
class DevEdit(FiggyTest):
    _VALUE = 'asdf'
    _DESC = 'desc'

    def __init__(self, extra_args=""):
        super().__init__(None, extra_args=extra_args)

    def run(self):
        self.step(f"Testing edit for {param_1}")
        self.edit()

    def edit(self):
        # Get Value
        child = pexpect.spawn(f'{CLI_NAME} config {Utils.get_first(edit)} --env {DEFAULT_ENV} '
                              f'--skip-upgrade {self.extra_args}',
                              timeout=10, encoding='utf-8')

        child.expect('.*Please input a PS Name.*')
        child.sendline(param_1)
        time.sleep(3) # Give edit time to start
        child.send(DevEdit._VALUE)
        child.sendcontrol('n')  # <-- sends TAB
        child.send(DevEdit._DESC)
        child.sendcontrol('n')  # <-- sends TAB
        child.sendcontrol('m')  # <-- Sends ENTER
        child.expect('.*secret.*')
        child.sendline('n')
        print("Add success. Checking successful save")

        get = DevGet(extra_args=self.extra_args)
        get.get(param_1, DevEdit._VALUE, DevEdit._DESC)
        delete = DevDelete(extra_args=self.extra_args)
        delete.delete(param_1)



import pexpect
import inspect
import re
import sys
from figcli.test.cli.base.restore import Restore


class RestoreTestSuite:

    @staticmethod
    def test_valid_role_key():
        sys.stdout.write("Testing single restore with valid role, valid key: ")

        restore = Restore(role="devops", env="dev")

        restore.choose_key(
            key="/devops/jordan/test/1", expect="Which item.+: "
        )

        restore.choose_restore_item(item_choice="2", expect=".+have it be the latest.+: ")

        restore.confirm_restore_item(".+Restore was successful\r\n")

    @staticmethod
    def test_valid_role_invalid_key():
        sys.stdout.write("Testing single restore with valid role, invalid key: ")

        restore = Restore(role="devops", env="dev")

        restore.choose_key(
            key="/devops/jordan/test/10",
            expect="No restorable values were found for this parameter.\r\n",
        )


def run():
    methods = inspect.getmembers(RestoreTestSuite, predicate=inspect.isfunction)

    for index, method in enumerate(methods):
        try:
            sys.stdout.write(f"Test #{index}: ")
            method[1]()
            sys.stdout.write("PASS\r\n")
        except pexpect.exceptions.EOF as eof:
            expected = re.findall("re.compile\(b'(.+)'\)", eof.value)[0]
            actual = re.findall("before \(last 100 chars\):.+'(.+)'", eof.value)[0]

            sys.stdout.write("FAIL\r\n")
            print(f"Expected => {expected}\r\nActual => {actual}")
        finally:
            sys.stdout.flush()


if __name__ == "__main__":
    run()